# Copyright (C) 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
import itertools
from os import remove
from pathlib import Path

import pytest
import yaml

from otx.api.entities.label import Domain
from otx.api.entities.model_template import (
    ANOMALY_TASK_TYPES,
    TRAINABLE_TASK_TYPES,
    DatasetRequirements,
    Dependency,
    EntryPoints,
    ExportableCodePaths,
    HyperParameterData,
    InstantiationType,
    ModelOptimizationMethod,
    ModelTemplate,
    NullModelTemplate,
    TargetDevice,
    TaskFamily,
    TaskType,
    _parse_model_template_from_omegaconf,
    parse_model_template,
    parse_model_template_from_dict,
    task_type_to_label_domain,
)
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


class CommonMethods:
    @staticmethod
    def cut_parameter_overrides_from_model_template(save_file: bool = True) -> dict:
        """Function saves config file with removed override_parameters node. Returns dictionary with path
        to new config file and dictionary with override_parameters"""
        with open(TestHyperParameterData().model_template_path()) as model_to_copy:
            model_data = yaml.safe_load(model_to_copy)
            override_parameters = model_data["hyper_parameters"]["parameter_overrides"]
        model_data["hyper_parameters"].pop("parameter_overrides")
        new_config_path = TestHyperParameterData.get_path_to_file(r"./no_overrides_template.yaml")
        if save_file:
            with open(new_config_path, "w+") as new_config:
                new_config.write(yaml.dump(model_data))
        return {
            "override_parameters": override_parameters,
            "new_config_path": new_config_path,
        }

    @staticmethod
    def check_model_attributes(model: ModelTemplate, expected_values: dict):
        assert model.model_template_id == expected_values.get("model_template_id")
        assert model.model_template_path == expected_values.get("model_template_path")
        assert model.name == expected_values.get("name")
        assert model.task_family == expected_values.get("task_family")
        assert model.task_type == expected_values.get("task_type")
        assert model.instantiation == expected_values.get("instantiation")
        assert model.summary == expected_values.get("summary", "")
        assert model.framework == expected_values.get("framework")
        assert model.max_nodes == expected_values.get("max_nodes", 1)
        assert model.application == expected_values.get("application")
        assert model.dependencies == expected_values.get("dependencies", [])
        assert model.initial_weights == expected_values.get("initial_weights")
        assert model.training_targets == expected_values.get("training_targets", [])
        assert model.inference_targets == expected_values.get("inference_targets", [])
        assert model.dataset_requirements == expected_values.get(
            "dataset_requirements", DatasetRequirements(classes=None)
        )
        assert model.model_optimization_methods == expected_values.get("model_optimization_methods", [])
        assert model.hyper_parameters == expected_values.get(
            "hyper_parameters",
            HyperParameterData(base_path=None, parameter_overrides={}),
        )
        assert model.is_trainable == expected_values.get("is_trainable", True)
        assert model.capabilities == expected_values.get("capabilities", [])
        assert model.grpc_address == expected_values.get("grpc_address")
        assert model.entrypoints == expected_values.get("entrypoints")
        assert model.exportable_code_paths == expected_values.get(
            "exportable_code_paths", ExportableCodePaths(default=None, openvino=None)
        )
        assert model.task_type_sort_priority == expected_values.get("task_type_sort_priority", -1)
        assert model.gigaflops == expected_values.get("gigaflops", 0)
        assert model.size == expected_values.get("size", 0)


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestTargetDevice:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_targetdevice(self):
        """
        <b>Description:</b>
        Check TargetDevice IntEnum class elements
        <b>Expected results:</b>
        Test passes if TargetDevice IntEnum class length is equal to expected value and its elements have expected
        sequence number
        """
        assert len(TargetDevice) == 4
        assert TargetDevice.UNSPECIFIED.value == 1
        assert TargetDevice.CPU.value == 2
        assert TargetDevice.GPU.value == 3
        assert TargetDevice.VPU.value == 4


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestModelOptimizationMethod:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_modeloptymizationmethod(self):
        """
        <b>Description:</b>
        Check ModelOptimizationMethod Enum class elements
        <b>Expected results:</b>
        Test passes if ModelOptimizationMethod Enum class length, methods and attributes return expected values
        <b>Steps</b>
        1. Check ModelOptimizationMethod length
        2. Check ModelOptimizationMethod elements value attribute
        3. Check ModelOptimizationMethod str method
        """
        assert len(ModelOptimizationMethod) == 2
        assert ModelOptimizationMethod.TENSORRT.value == 1
        assert ModelOptimizationMethod.OPENVINO.value == 2
        assert str(ModelOptimizationMethod.TENSORRT) == "TENSORRT"
        assert str(ModelOptimizationMethod.OPENVINO) == "OPENVINO"


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestDatasetRequirements:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_datasetrequirements(self):
        """
        <b>Description:</b>
        Check DatasetRequirements dataclass
        <b>Expected results:</b>
        Test passes if classes attribute of DatasetRequirements dataclass returns expected values
        """
        classes_list = ["class_1", "class_2"]
        test_dataset_requirements = DatasetRequirements(classes_list)
        equal_dataset_requirements = DatasetRequirements(classes_list)
        other_test_dataset_requirements = DatasetRequirements(["class_1", "class_3"])
        assert test_dataset_requirements.classes == ["class_1", "class_2"]
        assert other_test_dataset_requirements.classes == ["class_1", "class_3"]
        assert test_dataset_requirements == equal_dataset_requirements
        assert test_dataset_requirements != other_test_dataset_requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestExportableCodePaths:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_exportablecodepaths(self):
        """
        <b>Description:</b>
        Check ExportableCodePaths dataclass
        <b>Expected results:</b>
        Test passes if default, openvino attributes of ExportableCodePaths dataclass return expected values
        """
        exportable_code_paths = ExportableCodePaths("default code path", "openvino code path")
        equal_exportable_code_paths = ExportableCodePaths("default code path", "openvino code path")
        unequal_exportable_code_paths = ExportableCodePaths("other default code path", "openvino code path")
        assert exportable_code_paths.default == "default code path"
        assert exportable_code_paths.openvino == "openvino code path"
        assert exportable_code_paths == equal_exportable_code_paths
        assert exportable_code_paths != unequal_exportable_code_paths


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestTaskFamily:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_taskfamily(self):
        """
        <b>Description:</b>
        Check TaskFamily Enum class elements
        <b>Expected results:</b>
        Test passes if TaskFamily Enum class length, attributes and methods return expected values
        <b>Steps</b>
        1. Check TaskFamily length
        2. Check TaskFamily elements value attribute
        3. Check TaskFamily str method
        """
        assert len(TaskFamily) == 3
        assert TaskFamily.VISION.value == 1
        assert TaskFamily.FLOW_CONTROL.value == 2
        assert TaskFamily.DATASET.value == 3
        assert str(TaskFamily.VISION) == "VISION"
        assert str(TaskFamily.FLOW_CONTROL) == "FLOW_CONTROL"
        assert str(TaskFamily.DATASET) == "DATASET"


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestTaskType:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_tasktype(self):
        """
        <b>Description:</b>
        Check TaskType Enum class elements
        <b>Expected results:</b>
        Test passes if TaskType Enum class length, attributes and methods return expected values
        <b>Steps</b>
        1. Check TaskType length
        2. Check TaskType elements value attribute
        3. Check TaskType str method
        """
        assert len(TaskType) == 15
        assert TaskType.NULL.value == 1
        assert TaskType.DATASET.value == 2
        assert TaskType.CLASSIFICATION.value == 3
        assert TaskType.SEGMENTATION.value == 4
        assert TaskType.DETECTION.value == 5
        assert TaskType.ANOMALY_DETECTION.value == 6
        assert TaskType.CROP.value == 7
        assert TaskType.TILE.value == 8
        assert TaskType.INSTANCE_SEGMENTATION.value == 9
        assert TaskType.ACTIVELEARNING.value == 10
        assert TaskType.ANOMALY_SEGMENTATION.value == 11
        assert TaskType.ANOMALY_CLASSIFICATION.value == 12
        assert TaskType.ROTATED_DETECTION.value == 13
        assert TaskType.ACTION_CLASSIFICATION.value == 14
        assert TaskType.ACTION_DETECTION.value == 15
        assert str(TaskType.NULL) == "NULL"
        assert str(TaskType.DATASET) == "DATASET"
        assert str(TaskType.CLASSIFICATION) == "CLASSIFICATION"
        assert str(TaskType.SEGMENTATION) == "SEGMENTATION"
        assert str(TaskType.DETECTION) == "DETECTION"
        assert str(TaskType.ANOMALY_DETECTION) == "ANOMALY_DETECTION"
        assert str(TaskType.CROP) == "CROP"
        assert str(TaskType.TILE) == "TILE"
        assert str(TaskType.INSTANCE_SEGMENTATION) == "INSTANCE_SEGMENTATION"
        assert str(TaskType.ACTIVELEARNING) == "ACTIVELEARNING"
        assert str(TaskType.ANOMALY_SEGMENTATION) == "ANOMALY_SEGMENTATION"
        assert str(TaskType.ANOMALY_CLASSIFICATION) == "ANOMALY_CLASSIFICATION"
        assert str(TaskType.ROTATED_DETECTION) == "ROTATED_DETECTION"
        assert str(TaskType.ACTION_CLASSIFICATION) == "ACTION_CLASSIFICATION"
        assert str(TaskType.ACTION_DETECTION) == "ACTION_DETECTION"

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_tasktype_to_label_domain(self):
        """
        <b>Description:</b>
        Check task_type_to_label_domain function
        <b>Expected results:</b>
        Test passes if task_type_to_label_domain function returns expected mapping of TaskType class element
        to Domain class element
        expected values
        <b>Steps</b>
        1. Check positive scenario with existing TaskType element
        2. Check ValueError exception raised when requests non-existing TaskType element
        """
        assert task_type_to_label_domain(TaskType.CLASSIFICATION) == Domain.CLASSIFICATION
        assert task_type_to_label_domain(TaskType.DETECTION) == Domain.DETECTION
        assert task_type_to_label_domain(TaskType.SEGMENTATION) == Domain.SEGMENTATION
        assert task_type_to_label_domain(TaskType.INSTANCE_SEGMENTATION) == Domain.INSTANCE_SEGMENTATION
        assert task_type_to_label_domain(TaskType.ANOMALY_CLASSIFICATION) == Domain.ANOMALY_CLASSIFICATION
        assert task_type_to_label_domain(TaskType.ANOMALY_DETECTION) == Domain.ANOMALY_DETECTION
        assert task_type_to_label_domain(TaskType.ANOMALY_SEGMENTATION) == Domain.ANOMALY_SEGMENTATION
        for not_mapped_task in [
            TaskType.NULL,
            TaskType.DATASET,
            TaskType.CROP,
            TaskType.TILE,
            TaskType.ACTIVELEARNING,
            "key",
        ]:
            with pytest.raises(ValueError):
                task_type_to_label_domain(not_mapped_task)


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestHyperParameterData:
    @staticmethod
    def get_path_to_file(filename: str) -> str:
        """Return the path to the file named 'filename', which saved in tests/entities directory"""
        return str(Path(__file__).parent / Path(filename))

    def config_path(self) -> str:
        return self.get_path_to_file(r"./dummy_config.yaml")

    def model_template_path(self) -> str:
        return self.get_path_to_file(r"./dummy_template.yaml")

    @staticmethod
    def parameter_overrides() -> dict:
        return {
            "learning_parameters": {
                "batch_size": {"default_value": 10},
                "learning_rate": {"default_value": 0.5},
                "learning_rate_warmup_iters": {"min_value": 10},
                "num_checkpoints": {"max_value": 95},
                "num_iters": {"min_value": 2},
                "num_workers": {"description": "New workers description"},
            },
            "postprocessing": {
                "confidence_threshold": {"default_value": 0.4},
                "result_based_confidence_threshold": {"header": "New header"},
            },
        }

    def remove_value_key_from_config(self, config_content: dict) -> None:
        """Function removes "value" key from config dictionary"""
        config_content_copy = copy.deepcopy(config_content)
        for key, value in config_content_copy.items():
            if isinstance(value, dict):
                if key != "ui_rules":
                    self.remove_value_key_from_config(config_content[key])
            elif key == "value":
                config_content.pop(key)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_hyperparameterdata_data(self):
        """
        <b>Description:</b>
        Check data property of HyperParameterData class instance
        <b>Expected results:</b>
        Test passes if data property of HyperParameterData class instance has expected values
        <b>Steps</b>
        1. Check that data attribute of HyperParameterData object has empty dictionary before using load_parameters
        method
        2. Check scenario when path to config.yaml file specified in base_path parameter during HyperParameterData
        object initiation
        """

        def open_config_file_and_remove_value_key() -> dict:
            """Function returns config file dictionary with removed value key"""
            with open(TestHyperParameterData().config_path()) as config_file:
                config_content = yaml.safe_load(config_file)
                self.remove_value_key_from_config(config_content)
                return config_content

        # Forming expected data dictionary
        expected_config_data = open_config_file_and_remove_value_key()
        # Creating model template with no override_params_node
        model_template = CommonMethods.cut_parameter_overrides_from_model_template()
        model_template_path = model_template.get("new_config_path")
        # Checking data attribute of HyperParameterData class instance
        hyper_parameter_data = HyperParameterData(base_path=self.config_path())
        assert hyper_parameter_data.data == {}
        hyper_parameter_data.load_parameters(model_template_path)
        assert hyper_parameter_data.data == expected_config_data
        remove(model_template_path)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_remove_parameter_values_from_data(self):
        """
        <b>Description:</b>
        Check remove_parameter_values_from_data method of HyperParameterData class instance
        <b>Expected results:</b>
        Test passes if data property of HyperParameterData class instance has no "value" key after using
        load_parameters method
        <b>Steps</b>
        1. Check that yaml config file has "value" keys
        2. Check that data property of HyperParameterData class instance has no value key after using
        load_parameters method
        """

        def search_value_key_in_config(config_content: dict) -> bool:
            """
            Function returns "True" if "value" key presents in config dictionary and "False" if absents
            """
            is_value_key_exists = False
            config_content_copy = copy.deepcopy(config_content)
            for key, value in config_content_copy.items():
                if isinstance(value, dict):
                    if key != "ui_rules":
                        if search_value_key_in_config(config_content[key]):
                            is_value_key_exists = True
                            break
                elif key == "value":
                    is_value_key_exists = True
            return is_value_key_exists

        with open(self.config_path()) as config_file:
            config_data = yaml.safe_load(config_file)
        # Checking test dataset
        assert search_value_key_in_config(config_data)
        model_template = CommonMethods.cut_parameter_overrides_from_model_template()
        model_template_path = model_template.get("new_config_path")
        hyper_parameter_data = HyperParameterData(base_path=self.config_path())
        hyper_parameter_data.load_parameters(model_template_path)
        assert not search_value_key_in_config(hyper_parameter_data.data)
        remove(model_template_path)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_hyperparameterdata_has_overrides(self):
        """
        <b>Description:</b>
        Check has_overrides property of HyperParameterData class instance
        <b>Expected results:</b>
        Test passes if has_overrides property of HyperParameterData class instance has expected values
        <b>Steps</b>
        1. Check scenario when HyperParameterData object has specified parameter_overrides dictionary
        2. Check scenario when parameter_overrides dictionary not specified for HyperParameterData object
        """
        model_template = CommonMethods.cut_parameter_overrides_from_model_template()
        model_template_path = model_template.get("new_config_path")
        hyper_parameter_data = HyperParameterData(
            base_path=self.config_path(),
            parameter_overrides=model_template.get("override_parameters"),
        )
        hyper_parameter_data.load_parameters(model_template_path)
        has_overrides_hyper_parameter_data = HyperParameterData(
            base_path=self.config_path(), parameter_overrides=self.parameter_overrides()
        )
        assert has_overrides_hyper_parameter_data.has_overrides
        has_overrides_hyper_parameter_data.load_parameters(self.model_template_path())
        assert has_overrides_hyper_parameter_data.has_overrides
        no_overrides_hyper_parameter_data = HyperParameterData(base_path=self.config_path())
        assert not no_overrides_hyper_parameter_data.has_overrides
        no_overrides_hyper_parameter_data.load_parameters(model_template_path)
        assert not no_overrides_hyper_parameter_data.has_overrides
        remove(model_template_path)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_hyperparameterdata_has_valid_configurable_parameters(self):
        """
        <b>Description:</b>
        Check has_valid_configurable_parameters property of HyperParameterData class instance
        <b>Expected results:</b>
        Test passes if has_valid_configurable_parameters property of HyperParameterData class instance has
        expected values
        <b>Steps</b>
        1. Check scenario when base_path parameter specified during HyperParameterData object initiation and
        model_template_path file specified in load_parameters method exists
        2. Check scenario when __has_valid_configurable_parameters returns False when base_path parameter
        not specified during HyperParameterData object initiation
        3. Check scenario when __has_valid_configurable_parameters returns False when model_template_path file
        specified in load_parameters method not exists
        4. Check that ValueError exception raised when base_path file has unexpected structure
        """
        # positive scenario when expected has_valid_configurable_parameters = True
        model_template = CommonMethods.cut_parameter_overrides_from_model_template()
        model_template_path = model_template.get("new_config_path")
        hyper_parameter_data = HyperParameterData(base_path=self.config_path())
        hyper_parameter_data.load_parameters(model_template_path)
        assert hyper_parameter_data.has_valid_configurable_parameters
        # scenario when base_path parameter not specified
        no_config_specified_data = HyperParameterData()
        no_config_specified_data.load_parameters(model_template_path)
        assert not no_config_specified_data.base_path
        assert not no_config_specified_data.has_valid_configurable_parameters
        # scenario when model_template_path parameter not exists
        config_path = self.config_path()
        no_model_specified_data = HyperParameterData(config_path)
        no_model_specified_data.load_parameters(r"./file_not_exists.yaml")
        assert no_model_specified_data.base_path == config_path
        assert not no_config_specified_data.has_valid_configurable_parameters
        # check for incorrect config file
        incorrect_config_yaml_path = self.get_path_to_file(r"./incorrect_config.yaml")
        with open(incorrect_config_yaml_path, "w+") as incorrect_yaml_file:
            incorrect_yaml_file.write("[]")
        incorrect_config_data = HyperParameterData(incorrect_config_yaml_path)
        with pytest.raises(ValueError):
            incorrect_config_data.load_parameters(model_template_path)
        remove(model_template_path)
        remove(incorrect_config_yaml_path)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_hyperparameterdata_substitute_parameter_overrides(self):
        """
        <b>Description:</b>
        Check substitute_parameter_overrides method of HyperParameterData class instance
        <b>Expected results:</b>
        Test passes if substitute_parameter_overrides method returns data attribute of HyperParameterData class
        instance has expected value
        <b>Steps</b>
        1. Check positive scenario with valid parameter_overrides dictionary
        2. Check negative scenario with unexpected key to override
        3. Check negative scenario with "value" key to override in parameter_overrides dictionary
        """
        # positive scenario with valid overrides parameters dictionary
        model_template = CommonMethods.cut_parameter_overrides_from_model_template()
        model_template_path = model_template.get("new_config_path")
        hyper_parameter_data = HyperParameterData(
            base_path=self.config_path(), parameter_overrides=self.parameter_overrides()
        )
        hyper_parameter_data.load_parameters(model_template_path)
        learning_parameters_data = hyper_parameter_data.data.get("learning_parameters")
        postprocessing_data = hyper_parameter_data.data.get("postprocessing")
        assert learning_parameters_data.get("batch_size").get("default_value") == 10
        assert learning_parameters_data.get("learning_rate").get("default_value") == 0.5
        assert learning_parameters_data.get("learning_rate_warmup_iters").get("min_value") == 10
        assert learning_parameters_data.get("num_checkpoints").get("max_value") == 95
        assert learning_parameters_data.get("num_iters").get("min_value") == 2
        assert learning_parameters_data.get("num_workers").get("description") == "New workers description"
        assert postprocessing_data.get("confidence_threshold").get("default_value") == 0.4
        assert postprocessing_data.get("result_based_confidence_threshold").get("header") == "New header"
        # negative scenario with key not specified in config.yaml file
        unexpected_key_dict = {
            "learning_parameters": {"batch_size": {"default_value": 10}},
            "unexpected_key": {"parameter1": 1},
        }
        hyper_parameter_data = HyperParameterData(base_path=self.config_path(), parameter_overrides=unexpected_key_dict)
        with pytest.raises(ValueError):
            hyper_parameter_data.load_parameters(self.model_template_path())
        # negative scenario with "value" key not allowed to override
        restricted_key_dict = {"learning_parameters": {"batch_size": {"default_value": 10, "value": 1}}}
        hyper_parameter_data = HyperParameterData(base_path=self.config_path(), parameter_overrides=restricted_key_dict)
        with pytest.raises(KeyError):
            hyper_parameter_data.load_parameters(self.model_template_path())

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_manually_set_data_and_validate(self):
        """
        <b>Description:</b>
        Check manually_set_data_and_validate method of HyperParameterData class instance
        <b>Expected results:</b>
        Test passes if manually_set_data_and_validate method of HyperParameterData class instance has expected values
        <b>Steps</b>
        1. Check scenario when manually_set_data_and_validate method overrides HyperParameterData class data
        2. Check scenario when manually_set_data_and_validate method sets HyperParameterData class data
        3. Check scenario when manually_set_data_and_validate method removes HyperParameterData class data
        """
        # Check for override class data
        model_template = CommonMethods.cut_parameter_overrides_from_model_template()
        model_template_path = model_template.get("new_config_path")
        hyper_parameter_data = HyperParameterData(base_path=self.config_path())
        hyper_parameter_data.load_parameters(model_template_path)
        parameter_overrides = self.parameter_overrides()
        hyper_parameter_data.manually_set_data_and_validate(parameter_overrides)
        assert hyper_parameter_data.data == parameter_overrides
        assert hyper_parameter_data.has_valid_configurable_parameters
        # Check for set class data
        with open(self.config_path()) as config_file:
            data_to_set = yaml.safe_load(config_file)
        set_hyper_parameter_data = HyperParameterData(self.config_path())
        set_hyper_parameter_data.manually_set_data_and_validate(data_to_set)
        assert set_hyper_parameter_data.data == data_to_set
        assert set_hyper_parameter_data.has_valid_configurable_parameters
        # Check for set empty class data
        empty_hyper_parameter_data = HyperParameterData(self.config_path())
        empty_hyper_parameter_data.manually_set_data_and_validate({})
        assert empty_hyper_parameter_data.data == {}
        assert empty_hyper_parameter_data.has_valid_configurable_parameters
        remove(model_template_path)


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestInstantiationType:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_instantiationtype(self):
        """
        <b>Description:</b>
        Check InstantiationType Enum class elements
        <b>Expected results:</b>
        Test passes if InstantiationType Enum class length, methods and attributes return expected values
        <b>Steps</b>
        1. Check InstantiationType length
        2. Check InstantiationType elements value attribute
        3. Check InstantiationType str method
        """
        assert len(InstantiationType) == 3
        assert InstantiationType.NONE.value == 1
        assert InstantiationType.CLASS.value == 2
        assert InstantiationType.GRPC.value == 3
        assert str(InstantiationType.NONE) == "NONE"
        assert str(InstantiationType.CLASS) == "CLASS"
        assert str(InstantiationType.GRPC) == "GRPC"


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestDependency:
    @staticmethod
    def dependency_parameters() -> dict:
        return {
            "source": "dependency source",
            "destination": "dependency destination",
            "size": 1024,
            "sha256": "wgyjp0xks3obuwiu0jqea3q94ninfo0dgphv2t57wv1tq6qlwtr3jitvn0uo8a14",
        }

    def dependency(self) -> Dependency:
        return Dependency(**self.dependency_parameters())

    @staticmethod
    def unequal_dependency_parameters() -> dict:
        return {
            "source": "other source",
            "destination": "other destination",
            "size": 512,
            "sha256": "q0y3vrh9wcff5lc77epfx08pb6ioredv341u68rp1qvxyl41wzt9tlih94s5273i",
        }

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_depencdency(self):
        """
        <b>Description:</b>
        Check Dependency dataclass elements
        <b>Expected results:</b>
        Test passes if attributes of Dependency dataclass return expected values
        <b>Steps</b>
        1. Check Dependency source, destination, size and sha256 attributes of initiated Dependency object
        2. Check comparing Dependency object with equal object
        3. Check comparing Dependency object with unequal object
        """
        # Checking Dependency object attributes
        expected_source = self.dependency_parameters().get("source")
        expected_destination = self.dependency_parameters().get("destination")
        dependency = self.dependency()
        assert dependency.source == expected_source
        assert dependency.destination == expected_destination
        assert dependency.size == self.dependency_parameters().get("size")
        assert dependency.sha256 == self.dependency_parameters().get("sha256")
        default_attributes_dependency = Dependency(source=expected_source, destination=expected_destination)
        assert default_attributes_dependency.source == expected_source
        assert default_attributes_dependency.destination == expected_destination
        assert not default_attributes_dependency.size
        assert not default_attributes_dependency.sha256
        # Comparing Dependency object with equal
        equal_dependency = self.dependency()
        assert dependency == equal_dependency
        # Comparing Dependency object with unequal
        keys_list = ["source", "destination", "size", "sha256"]
        parameter_combinations = []
        for i in range(1, len(keys_list) + 1):
            parameter_combinations.append(list(itertools.combinations(keys_list, i)))
        # In each of scenario creating a copy of equal parameters and replacing to values from prepared
        # dictionary
        unequal_dependency_parameters = self.unequal_dependency_parameters()
        for scenario in parameter_combinations:
            for parameters in scenario:
                unequal_params_dict = dict(self.dependency_parameters())
                for key in parameters:
                    unequal_params_dict[key] = unequal_dependency_parameters.get(key)
                unequal_dependency = Dependency(**unequal_params_dict)
                assert dependency != unequal_dependency, (
                    "Failed to check that Dependency instances with different " f"{parameters} are unequal"
                )


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestEntryPoints:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_entrypoints(self):
        """
        <b>Description:</b>
        Check EntryPoints dataclass elements
        <b>Expected results:</b>
        Test passes if classes attributes of EntryPoints dataclass return expected values
        <b>Steps</b>
        1. Check EntryPoints base, openvino and nncf attributes of initiated Dependency object
        2. Check comparing EntryPoints object with equal object
        3. Check comparing EntryPoints object with unequal object
        """
        base = "base interface"
        openvino = "OpenVINO interface"
        nncf = "NNCF interface"
        entrypoints_parameters = {"base": base, "openvino": openvino, "nncf": nncf}
        # Checking EntryPoints object attributes
        entry_points = EntryPoints(**entrypoints_parameters)
        assert entry_points.base == base
        assert entry_points.openvino == openvino
        assert entry_points.nncf == nncf
        default_attributes_entrypoints = EntryPoints(base=base)
        assert default_attributes_entrypoints.base == base
        assert not default_attributes_entrypoints.openvino
        assert not default_attributes_entrypoints.nncf
        # Comparing EntryPoints object with equal
        equal_entry_points = EntryPoints(**entrypoints_parameters)
        assert entry_points == equal_entry_points
        # Comparing Dependency object with unequal
        keys_list = ["base", "openvino", "nncf"]
        parameter_combinations = []
        for i in range(1, len(keys_list) + 1):
            parameter_combinations.append(list(itertools.combinations(keys_list, i)))
        # In each of scenario creating a copy of equal parameters and replacing to values from prepared
        # dictionary
        unequal_entrypoints_parameters = {
            "base": "other base interface",
            "openvino": "other OpenVINO interface",
            "nncf": "other NNCF interface",
        }
        for scenario in parameter_combinations:
            for parameters in scenario:
                unequal_params_dict = dict(entrypoints_parameters)
                for key in parameters:
                    unequal_params_dict[key] = unequal_entrypoints_parameters.get(key)
                unequal_entry_points = EntryPoints(**unequal_params_dict)
                assert entry_points != unequal_entry_points, (
                    "Failed to check that EntryPoints instances with " f"different {parameters} are unequal"
                )


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestModelTemplate:
    @staticmethod
    def default_model_parameters() -> dict:
        return {
            "model_template_id": "A16",
            "model_template_path": TestHyperParameterData().model_template_path(),
            "name": "test_model",
            "task_family": TaskFamily.DATASET,
            "task_type": TaskType.DETECTION,
            "instantiation": InstantiationType.NONE,
        }

    def optional_model_parameters(self):
        model_template = CommonMethods.cut_parameter_overrides_from_model_template()
        model_template_path = model_template.get("new_config_path")
        hyper_parameter_data = HyperParameterData(base_path=TestHyperParameterData().config_path())
        hyper_parameter_data.load_parameters(model_template_path)
        optional_parameters = dict(self.default_model_parameters())
        optional_parameters["model_template_path"] = model_template_path
        optional_parameters["model_template_id"] = "B18"
        optional_parameters["name"] = "test_model2"
        optional_parameters["task_family"] = TaskFamily.VISION
        optional_parameters["summary"] = "algorithm related information"
        optional_parameters["framework"] = "test framework"
        optional_parameters["max_nodes"] = 2
        optional_parameters["application"] = "test application"
        optional_parameters["dependencies"] = [
            TestDependency().dependency(),
            Dependency(**TestDependency().unequal_dependency_parameters()),
        ]
        optional_parameters["initial_weights"] = "https://some_url.com"
        optional_parameters["training_targets"] = [
            TargetDevice.CPU,
            TargetDevice.GPU,
            TargetDevice.VPU,
        ]
        optional_parameters["inference_targets"] = [
            TargetDevice.CPU,
            TargetDevice.VPU,
            TargetDevice.GPU,
        ]
        optional_parameters["dataset_requirements"] = [
            DatasetRequirements(["class1", "class_2"]),
            DatasetRequirements(["class_3", "class_4"]),
        ]
        optional_parameters["model_optimization_methods"] = [
            ModelOptimizationMethod.OPENVINO,
            ModelOptimizationMethod.TENSORRT,
        ]
        optional_parameters["hyper_parameters"] = hyper_parameter_data
        optional_parameters["is_trainable"] = False
        optional_parameters["capabilities"] = [
            "compute_uncertainty_score",
            "compute_representations",
        ]
        optional_parameters["grpc_address"] = "192.168.1.1"
        optional_parameters["entrypoints"] = EntryPoints.openvino
        optional_parameters["exportable_code_paths"] = ExportableCodePaths.openvino
        optional_parameters["task_type_sort_priority"] = 0
        optional_parameters["gigaflops"] = 1
        optional_parameters["size"] = 1024
        return optional_parameters

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_model_template_initiation(self):
        """
        <b>Description:</b>
        Check TestModelTemplate dataclass attributes
        <b>Expected results:</b>
        Test passes if classes attributes of ModelTemplate object return expected values
        <b>Steps</b>
        1. Check attributes of ModelTemplate object initiated with default values
        2. Check attributes of ModelTemplate object initiated with fully specified values
        """
        # Checks for object with default values
        default_model_template = ModelTemplate(**self.default_model_parameters())
        CommonMethods.check_model_attributes(default_model_template, self.default_model_parameters())
        # Checks for object with specified values
        optional_parameters_model = ModelTemplate(**self.optional_model_parameters())
        CommonMethods.check_model_attributes(optional_parameters_model, self.optional_model_parameters())
        assert default_model_template != optional_parameters_model
        remove(optional_parameters_model.model_template_path)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_model_template_post_initialization(self):
        """
        <b>Description:</b>
        Check __post_init__ method of ModelTemplate dataclass
        <b>Expected results:</b>
        Test passes if __post_init__ method of ModelTemplate object raises ValueError exception for parameters violation
        <b>Steps</b>
        1. Check ValueError exception when initialized ModelTemplate has gRPC instantiation and "" gRPC address
        2. Check ValueError exception when initialized ModelTemplate has CLASS instantiation and no entry points
        3. Check ValueError exception when initialized ModelTemplate has VISION task family and has no specified path to
        config file
        4. Check ValueError exception when initialized ModelTemplate with task family in not equal to VISION and but has
        specified path to config file
        """
        model_template_parameters = self.default_model_parameters()
        # gRPC instantiation and "" GRPC address
        grpc_address_empty = dict(model_template_parameters)
        grpc_address_empty["instantiation"] = InstantiationType.GRPC
        grpc_address_empty["grpc_address"] = ""
        with pytest.raises(ValueError):
            ModelTemplate(**grpc_address_empty)
        # CLASS instantiation and no entrypoints
        class_no_entrypoints = dict(model_template_parameters)
        class_no_entrypoints["instantiation"] = InstantiationType.CLASS
        with pytest.raises(ValueError):
            ModelTemplate(**class_no_entrypoints)
        # VISION task family no config file path specified
        vision_no_config = dict(model_template_parameters)
        vision_no_config["task_family"] = TaskFamily.VISION
        with pytest.raises(ValueError):
            ModelTemplate(**vision_no_config)
        # Not VISION task family and specified path to config
        class_not_vision_with_config = dict(model_template_parameters)
        class_not_vision_with_config["hyper_parameters"] = HyperParameterData(
            base_path=TestHyperParameterData().config_path()
        )
        with pytest.raises(ValueError):
            ModelTemplate(**class_not_vision_with_config)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_model_template_capabilities(self):
        """
        <b>Description:</b>
        Check computes_uncertainty_score and computes_representations methods of ModelTemplate dataclass
        <b>Expected results:</b>
        Test passes if computes_uncertainty_score and computes_representations methods of ModelTemplate object return
        expected bool values related to capabilities attribute
        <b>Steps</b>
        1. Check computes_uncertainty_score and computes_representations methods return True value for ModelTemplate
        2. Check computes_uncertainty_score method returns True and computes_representations returns False value
        3. Check computes_uncertainty_score method returns False and computes_representations returns True value
        4. Check computes_uncertainty_score and computes_representations methods return False value
        """
        # Check for computes_uncertainty_score and computes_representations methods return True
        score_representations_model = ModelTemplate(**self.optional_model_parameters())
        assert score_representations_model.computes_uncertainty_score()
        assert score_representations_model.computes_representations()
        model_template_parameters = self.default_model_parameters()
        # Check for computes_uncertainty_score is True and computes_representations is False
        score_true_presentations_false_parameters = dict(model_template_parameters)
        score_true_presentations_false_parameters["capabilities"] = [
            "compute_uncertainty_score",
            "not test parameter",
        ]
        score_true_presentations_false_model = ModelTemplate(**score_true_presentations_false_parameters)
        assert score_true_presentations_false_model.computes_uncertainty_score()
        assert not score_true_presentations_false_model.computes_representations()
        # Check for computes_uncertainty_score is False and computes_representations is True
        score_true_presentations_false_parameters = dict(model_template_parameters)
        score_true_presentations_false_parameters["capabilities"] = [
            "compute_representations",
            "not test parameter",
        ]
        score_true_presentations_false_model = ModelTemplate(**score_true_presentations_false_parameters)
        assert not score_true_presentations_false_model.computes_uncertainty_score()
        assert score_true_presentations_false_model.computes_representations()
        # Check for computes_uncertainty_score and computes_representations methods return False
        no_score_representations_model = ModelTemplate(**self.default_model_parameters())
        assert not no_score_representations_model.computes_uncertainty_score()
        assert not no_score_representations_model.computes_representations()
        remove(score_representations_model.model_template_path)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_model_template_is_task_global(self):
        """
        <b>Description:</b>
        Check is_task_global method of ModelTemplate dataclass
        <b>Expected results:</b>
        Test passes if is_task_global method of ModelTemplate object returns expected bool values related to
        task_type attribute
        <b>Steps</b>
        1. Check is_task_global method returns True if task_type equal to CLASSIFICATION or ANOMALY_CLASSIFICATION
        2. Check is_task_global method returns False if task_type not equal to CLASSIFICATION or ANOMALY_CLASSIFICATION
        """
        # Check is_task_global method returns True for CLASSIFICATION and ANOMALY_CLASSIFICATION
        for global_task_type in (
            TaskType.CLASSIFICATION,
            TaskType.ANOMALY_CLASSIFICATION,
        ):
            default_parameters = self.default_model_parameters()
            task_global_parameters = dict(default_parameters)
            task_global_parameters["task_type"] = global_task_type
            task_global_model_template = ModelTemplate(**task_global_parameters)
            assert (
                task_global_model_template.is_task_global()
            ), f"Expected True value returned by is_task_global for {global_task_type}"
        # Check is_task_global method returns False for the other tasks
        non_global_task_parameters = dict(default_parameters)
        non_global_tasks_list = []
        for task_type in TaskType:
            if not task_type.is_global:
                non_global_tasks_list.append(task_type)
        for non_global_task in non_global_tasks_list:
            non_global_task_parameters["task_type"] = non_global_task
            non_global_task_template = ModelTemplate(**non_global_task_parameters)
            assert not non_global_task_template.is_task_global(), (
                f"Expected False value returned by is_task_global method for {non_global_task}, "
                f"only CLASSIFICATION and ANOMALY_CLASSIFICATION task types are global"
            )


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestNullModelTemplate:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_null_mode_attributes(self):
        """
        <b>Description:</b>
        Check attributes of NullModelTemplate class object
        <b>Expected results:</b>
        Test passes if NullModelTemplate class object attributes have expected values
        """
        null_model_template = NullModelTemplate()
        expected_null_model_parameters = {
            "model_template_id": "",
            "model_template_path": "",
            "name": "Null algorithm",
            "task_family": TaskFamily.FLOW_CONTROL,
            "task_type": TaskType.NULL,
            "instantiation": InstantiationType.NONE,
        }
        CommonMethods.check_model_attributes(null_model_template, expected_null_model_parameters)


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestTaskTypesConstants:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_task_type_constants(self):
        """
        <b>Description:</b>
        Check values of ANOMALY_TASK_TYPES and TRAINABLE_TASK_TYPES constants
        <b>Expected results:</b>
        Test passes if ANOMALY_TASK_TYPES and TRAINABLE_TASK_TYPES constants return expected values
        """
        assert ANOMALY_TASK_TYPES == (
            TaskType.ANOMALY_DETECTION,
            TaskType.ANOMALY_CLASSIFICATION,
            TaskType.ANOMALY_SEGMENTATION,
        )
        assert TRAINABLE_TASK_TYPES == (
            TaskType.CLASSIFICATION,
            TaskType.DETECTION,
            TaskType.SEGMENTATION,
            TaskType.INSTANCE_SEGMENTATION,
            TaskType.ANOMALY_DETECTION,
            TaskType.ANOMALY_CLASSIFICATION,
            TaskType.ANOMALY_SEGMENTATION,
            TaskType.ROTATED_DETECTION,
            TaskType.ACTION_CLASSIFICATION,
            TaskType.ACTION_DETECTION,
        )


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestParseModelTemplate:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_parse_model_template_from_omegaconf(self):
        """
        <b>Description:</b>
        Check _parse_model_template_from_omegaconf function returns expected instance of ModelTemplate class
        <b>Expected results:</b>
        Test passes if _parse_model_template_from_omegaconf function returns expected instance of ModelTemplate
        class
        """
        model_template_path = TestHyperParameterData().model_template_path()
        with open(model_template_path) as model_template_file:
            model_template_content = yaml.safe_load(model_template_file)
            model_template_content["model_template_id"] = model_template_content["name"].replace(" ", "_")
            model_template_content["model_template_path"] = model_template_path
        parsed_model_template = _parse_model_template_from_omegaconf(model_template_content)
        assert isinstance(parsed_model_template, ModelTemplate)
        # Forming expected hyper parameter data dictionary from config.yaml and overridden parameters from
        # model_template,yaml
        parameter_overrides = {
            "learning_parameters": {
                "batch_size": {"default_value": 64},
                "learning_rate": {"default_value": 0.05},
                "learning_rate_warmup_iters": {"default_value": 100},
                "num_iters": {"default_value": 13000},
            }
        }
        expected_hyper_parameters = HyperParameterData(
            base_path="./dummy_config.yaml", parameter_overrides=parameter_overrides
        )
        expected_hyper_parameters.load_parameters(model_template_path)
        expected_parsed_model_parameters = {
            "entrypoints": EntryPoints(base="base entrypoints", openvino=None, nncf=None),
            "framework": "Test framework",
            "hyper_parameters": expected_hyper_parameters,
            "inference_targets": [
                TargetDevice.CPU,
                TargetDevice.GPU,
                TargetDevice.VPU,
            ],
            "instantiation": InstantiationType.CLASS,
            "model_template_id": "Custom_Object_Detection_--_TEST_ONLY",
            "model_template_path": model_template_path,
            "name": "Custom Object Detection -- TEST ONLY",
            "summary": "Fast and lightweight object detector.",
            "task_family": TaskFamily.VISION,
            "task_type": TaskType.DETECTION,
            "training_targets": [TargetDevice.GPU, TargetDevice.CPU],
        }
        CommonMethods.check_model_attributes(parsed_model_template, expected_parsed_model_parameters)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_parse_model_template(self):
        """
        <b>Description:</b>
        Check parse_model_template function returns expected instance of ModelTemplate class
        <b>Expected results:</b>
        Test passes if parse_model_template function returns instance of ModelTemplate class
        <b>Steps</b>
        1. Check model_template_id and model_template_path attributes of ModelTemplate instance returned by
        parse_model_template function for template file with not specified model_template_id parameter
        2. Check model_template_id and model_template_path attributes of ModelTemplate instance returned by
        parse_model_template function for template file with specified model_template_id parameter
        3. Check ValueError exception raised if path to list-type template file is specified as input parameter in
        parse_model_template function
        """
        # Check for template file with not specified model_template_id
        model_template_path = TestHyperParameterData().model_template_path()
        not_specified_id_template = parse_model_template(model_template_path)
        assert not_specified_id_template.model_template_id == "Custom_Object_Detection_--_TEST_ONLY"
        assert not_specified_id_template.model_template_path == model_template_path
        # Check for template file with specified model_template_id
        id_specified_model_path = TestHyperParameterData.get_path_to_file(r"./id_specified_template.yaml")
        model_id = "Parsed_Model_ID_1"
        with open(model_template_path) as model_template_file:
            id_specified_template_content = yaml.safe_load(model_template_file)
            id_specified_template_content["model_template_id"] = model_id
        with open(id_specified_model_path, "w") as new_template:
            new_template.write(yaml.dump(id_specified_template_content))
        id_specified_template = parse_model_template(id_specified_model_path)
        assert id_specified_template.model_template_id == model_id
        assert id_specified_template.model_template_path == id_specified_model_path
        remove(id_specified_model_path)
        # Check ValueError exception raised if model template is list-type
        incorrect_model_template_path = TestHyperParameterData.get_path_to_file(r"./incorrect_model_template.yaml")
        with open(incorrect_model_template_path, "w+") as incorrect_yaml_file:
            incorrect_yaml_file.write("[]")
        with pytest.raises(ValueError):
            parse_model_template(incorrect_model_template_path)
        remove(incorrect_model_template_path)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_parse_model_template_from_dict(self):
        """
        <b>Description:</b>
        Check parse_model_template_from_dict function returns expected instance of ModelTemplate class
        <b>Expected results:</b>
        Test passes if parse_model_template_from_dict function returns instance of ModelTemplate class
        <b>Steps</b>
        1. Check ModelTemplate instance returned by test_parse_model_template_from_dict function for dictionary
        with specified model_template_id and model_template_path
        parameters
        """
        model_template = CommonMethods.cut_parameter_overrides_from_model_template()
        model_template_path = model_template.get("new_config_path")
        override_parameters = model_template.get("override_parameters")
        hyper_parameters = HyperParameterData(
            base_path=TestHyperParameterData().config_path(),
            parameter_overrides=override_parameters,
        )
        hyper_parameters.load_parameters(model_template_path)
        template_dictionary = {
            "model_template_id": "Dictionary_Model_1",
            "model_template_path": model_template_path,
            "name": "Custom Object Detection -- TEST ONLY",
            "task_type": TaskType.DETECTION,
            "task_family": TaskFamily.VISION,
            "instantiation": InstantiationType.CLASS,
            "summary": "Fast and lightweight object detector.",
            "application": None,
            "framework": "Test framework",
            "entrypoints": EntryPoints(base="base interface", openvino="OpenVINO interface"),
            "hyper_parameters": hyper_parameters,
            "max_nodes": 1,
            "training_targets": [TargetDevice.GPU, TargetDevice.CPU],
            "inference_targets": [
                TargetDevice.CPU,
                TargetDevice.GPU,
                TargetDevice.VPU,
            ],
        }
        model_template_from_dictionary = parse_model_template_from_dict(template_dictionary)
        CommonMethods.check_model_attributes(model_template_from_dictionary, template_dictionary)
        remove(model_template_path)



"""OTX AsyncPipeline of OTX Detection."""

# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import copy
from time import perf_counter

from openvino.model_zoo.model_api.pipelines import AsyncPipeline


class OTXDetectionAsyncPipeline(AsyncPipeline):
    """OTX AsyncPipeline of OTX Detection."""

    def get_result(self, id):  # pylint: disable=redefined-builtin
        """Get result of inference by index.

        Args:
            id (int): index of inference

        Returns:
            result (tuple): tuple of inference result and meta information
        """
        result = self.get_raw_result(id)
        if result:
            raw_result, meta, preprocess_meta, infer_start_time = result
            self.inference_metrics.update(infer_start_time)

            postprocessing_start_time = perf_counter()
            result = self.model.postprocess(raw_result, preprocess_meta), {**meta, **preprocess_meta}
            self.postprocess_metrics.update(postprocessing_start_time)
            features = (None, None)
            if "feature_vector" in raw_result or "saliency_map" in raw_result:
                features = (
                    copy.deepcopy(raw_result["feature_vector"].reshape(-1)),
                    copy.deepcopy(raw_result["saliency_map"][0]),
                )
            return *result, features
        return None



"""Test MMSegmentationTask."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
import torch
import pytest
from mmcv import ConfigDict

from otx.algorithms.segmentation.adapters.mmseg.task import MMSegmentationTask
from otx.api.configuration.helper import create
from otx.api.entities.model_template import (
    parse_model_template,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.segmentation.test_helpers import (
    DEFAULT_SEG_TEMPLATE_DIR,
    init_environment,
)


class TestMMSegmentationTask:
    @pytest.fixture(autouse=True)
    def setup(self, mocker):
        model_template = parse_model_template(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "template.yaml"))
        hyper_parameters = create(model_template.hyper_parameters.data)
        task_env = init_environment(hyper_parameters, model_template)
        self.mmseg_task = MMSegmentationTask(task_env)
        self.mmseg_task._init_task()

    @e2e_pytest_unit
    def test_configure(self):
        cfg = self.mmseg_task.configure()
        assert "work_dir" in cfg
        assert "resume" in cfg

    @e2e_pytest_unit
    def test_build_model(self, mocker):
        mocker.patch("otx.algorithms.segmentation.adapters.mmseg.utils.builder.build_segmentor")
        self.mmseg_task._recipe_cfg.model.decode_head.type = "CustomFCNHead"
        self.mmseg_task._recipe_cfg.model.task_adapt = ConfigDict(
            op="REPLACE",
            type="mpa",
            final=["background", "target"],
            src_classes=[],
            dst_classes=["background", "target"],
        )
        model = self.mmseg_task.build_model(self.mmseg_task._recipe_cfg, fp16=True)
        assert isinstance(model, torch.nn.Module)
        assert model.fp16_enabled

    @e2e_pytest_unit
    def test_update_override_configurations(self):
        cfg = ConfigDict(fake_key="fake_value")
        self.mmseg_task.update_override_configurations(cfg)
        assert "fake_key" in self.mmseg_task.override_configs
        assert self.mmseg_task.override_configs == dict(fake_key="fake_value")

    @e2e_pytest_unit
    def test_save_model(self, otx_model, mocker):
        mocker_load = mocker.patch("torch.load", return_value="foo")
        mocker_save = mocker.patch("torch.save")
        self.mmseg_task.save_model(otx_model)

        mocker_load.assert_called_once()
        mocker_save.assert_called_once()



"""Dynamic Action classification mmdeply cfg."""

_base_ = ["./base_classification_static.py"]

ir_config = dict(
    dynamic_axes=dict(
        data=dict({0: "batch"}),
        logits=dict({0: "batch"}),
    ),
)



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest

from otx.algorithms.segmentation.adapters.mmseg.datasets.pipelines.loads import (
    LoadAnnotationFromOTXDataset,
)
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from tests.test_suite.e2e_test_system import e2e_pytest_unit


def label_entity(name="test label") -> LabelEntity:
    return LabelEntity(name=name, domain=Domain.SEGMENTATION)


def dataset_item() -> DatasetItemEntity:
    image: Image = Image(data=np.random.randint(low=0, high=255, size=(10, 16, 3)))
    annotation: Annotation = Annotation(shape=Rectangle.generate_full_box(), labels=[ScoredLabel(label_entity())])
    annotation_scene: AnnotationSceneEntity = AnnotationSceneEntity(
        annotations=[annotation], kind=AnnotationSceneKind.ANNOTATION
    )
    return DatasetItemEntity(media=image, annotation_scene=annotation_scene)


class TestLoadAnnotationFromOTXDataset:
    @pytest.fixture(autouse=True)
    def setUp(self) -> None:

        self.dataset_item: DatasetItemEntity = dataset_item()
        self.results: dict = {
            "dataset_item": self.dataset_item,
            "ann_info": {"labels": [label_entity("class_1")]},
            "seg_fields": [],
        }
        self.pipeline: LoadAnnotationFromOTXDataset = LoadAnnotationFromOTXDataset()

    @e2e_pytest_unit
    def test_call(self) -> None:
        loaded_annotations: dict = self.pipeline(self.results)
        assert "gt_semantic_seg" in loaded_annotations
        assert loaded_annotations["dataset_item"] == self.dataset_item



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest

from otx.api.configuration.enums.config_element_type import (
    ConfigElementType,
    ElementCategory,
)
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestElementCategory:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_element_category(self):
        """
        <b>Description:</b>
        Check "ElementCategory" Enum class elements

        <b>Expected results:</b>
        Test passes if "ElementCategory" Enum class length is equal to expected value and its elements have expected
        sequence numbers, "name" attributes and values returned by __str__ method
        """
        assert len(ElementCategory) == 3
        assert ElementCategory.PRIMITIVES.value == 1
        assert ElementCategory.PRIMITIVES.name == "PRIMITIVES"
        assert str(ElementCategory.PRIMITIVES) == "PRIMITIVES"
        assert ElementCategory.GROUPS.value == 2
        assert ElementCategory.GROUPS.name == "GROUPS"
        assert str(ElementCategory.GROUPS) == "GROUPS"
        assert ElementCategory.RULES.value == 3
        assert ElementCategory.RULES.name == "RULES"
        assert str(ElementCategory.RULES) == "RULES"


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestConfigElementType:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_element_category(self):
        """
        <b>Description:</b>
        Check "ConfigElementType" Enum class elements

        <b>Expected results:</b>
        Test passes if "ConfigElementType" Enum class length is equal to expected value and its elements have expected
        sequence numbers, "names" attributes, "category" properties and values returned by and __str__ method
        """
        assert len(ConfigElementType) == 9
        # Checking "INTEGER" element
        assert ConfigElementType.INTEGER.category == ElementCategory.PRIMITIVES
        assert ConfigElementType.INTEGER.name == "INTEGER"
        assert ConfigElementType.INTEGER.value == 0
        assert str(ConfigElementType.INTEGER) == "INTEGER"
        # Checking "FLOAT" element
        assert ConfigElementType.FLOAT.category == ElementCategory.PRIMITIVES
        assert ConfigElementType.FLOAT.name == "FLOAT"
        assert ConfigElementType.FLOAT.value == 1
        assert str(ConfigElementType.FLOAT) == "FLOAT"
        # Checking "BOOLEAN" element
        assert ConfigElementType.BOOLEAN.category == ElementCategory.PRIMITIVES
        assert ConfigElementType.BOOLEAN.name == "BOOLEAN"
        assert ConfigElementType.BOOLEAN.value == 2
        assert str(ConfigElementType.BOOLEAN) == "BOOLEAN"
        # Checking "FLOAT_SELECTABLE" element
        assert ConfigElementType.FLOAT_SELECTABLE.category == ElementCategory.PRIMITIVES
        assert ConfigElementType.FLOAT_SELECTABLE.name == "FLOAT_SELECTABLE"
        assert ConfigElementType.FLOAT_SELECTABLE.value == 3
        assert str(ConfigElementType.FLOAT_SELECTABLE) == "FLOAT_SELECTABLE"
        # Checking "SELECTABLE" element
        assert ConfigElementType.SELECTABLE.category == ElementCategory.PRIMITIVES
        assert ConfigElementType.SELECTABLE.name == "SELECTABLE"
        assert ConfigElementType.SELECTABLE.value == 4
        assert str(ConfigElementType.SELECTABLE) == "SELECTABLE"
        # Checking "PARAMETER_GROUP" element
        assert ConfigElementType.PARAMETER_GROUP.category == ElementCategory.GROUPS
        assert ConfigElementType.PARAMETER_GROUP.name == "PARAMETER_GROUP"
        assert ConfigElementType.PARAMETER_GROUP.value == 5
        assert str(ConfigElementType.PARAMETER_GROUP) == "PARAMETER_GROUP"
        # Checking "CONFIGURABLE_PARAMETERS" element
        assert ConfigElementType.CONFIGURABLE_PARAMETERS.category == ElementCategory.GROUPS
        assert ConfigElementType.CONFIGURABLE_PARAMETERS.name == "CONFIGURABLE_PARAMETERS"
        assert ConfigElementType.CONFIGURABLE_PARAMETERS.value == 6
        assert str(ConfigElementType.CONFIGURABLE_PARAMETERS) == "CONFIGURABLE_PARAMETERS"
        # Checking "RULE" element
        assert ConfigElementType.RULE.category == ElementCategory.RULES
        assert ConfigElementType.RULE.name == "RULE"
        assert ConfigElementType.RULE.value == 7
        assert str(ConfigElementType.RULE) == "RULE"
        # Checking "RULE" element
        assert ConfigElementType.UI_RULES.category == ElementCategory.RULES
        assert ConfigElementType.UI_RULES.name == "UI_RULES"
        assert ConfigElementType.UI_RULES.value == 8
        assert str(ConfigElementType.UI_RULES) == "UI_RULES"



import argparse

import pytest

from otx.cli.tools import build as target_package
from otx.cli.tools.build import get_args, main
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@e2e_pytest_unit
def test_get_args(mocker):
    mock_options = {
        "--train-data-roots": "train/data/root",
        "--val-data-roots": "val/data/root",
        "--test-data-roots": "test/data/root",
        "--unlabeled-data-roots": "unlabeled/data/root",
        "--unlabeled-file-list": "unlabeled/file/list",
        "--task": "detection",
        "--train-type": "Semisupervised",
        "--workspace": "work/dir/path",
        "--model": "SSD",
        "--backbone": "torchvision.resnet18",
    }
    mock_command = ["otx"]
    for key, value in mock_options.items():
        mock_command.extend([key, value])

    mocker.patch("sys.argv", mock_command)
    mocker.patch.object(target_package, "get_parser_and_hprams_data", return_value=[argparse.ArgumentParser(), {}, []])

    parsed_args = get_args()

    assert parsed_args.train_data_roots == "train/data/root"
    assert parsed_args.val_data_roots == "val/data/root"
    assert parsed_args.test_data_roots == "test/data/root"
    assert parsed_args.unlabeled_data_roots == "unlabeled/data/root"
    assert parsed_args.unlabeled_file_list == "unlabeled/file/list"
    assert parsed_args.workspace == "work/dir/path"
    assert parsed_args.task == "detection"
    assert parsed_args.train_type == "Semisupervised"
    assert parsed_args.model == "SSD"
    assert parsed_args.backbone == "torchvision.resnet18"


@pytest.fixture
def mock_args(mocker, tmp_path):
    mock_args = mocker.MagicMock()
    mock_args.train_data_roots = None
    mock_args.val_data_roots = None
    mock_args.test_data_roots = None
    mock_args.unlabeled_data_roots = None
    mock_args.unlabeled_file_list = None
    mock_args.task = ""
    mock_args.train_type = "incremental"
    mock_args.workspace = tmp_path / "work_dir"
    mock_args.model = ""
    mock_args.backbone = "torchvision.resnet18"

    def mock_contains(self, val):
        return val in self.__dict__

    mock_args.__contains__ = mock_contains
    mock_get_args = mocker.patch("otx.cli.tools.build.get_args")
    mock_get_args.return_value = mock_args

    return mock_args


@pytest.fixture
def mock_config_manager(mocker):
    mock_config_manager = mocker.patch.object(target_package, "ConfigManager")
    mock_template = mocker.MagicMock()
    mock_template.name = "fake_name"
    mock_config_manager.return_value.mode = "build"
    mock_config_manager.return_value.template = mock_template
    mock_config_manager.return_value.check_workspace.return_value = True
    mock_config_manager.return_value.get_dataset_config.return_value = {}
    mock_config_manager.return_value.get_hyparams_config.return_value = {}

    return mock_config_manager


@e2e_pytest_unit
def test_main(mocker, mock_config_manager, mock_args):
    # Mock argparse namespace
    mock_builder = mocker.patch("otx.cli.builder.Builder")

    # Call main function
    result = main()

    # Check return value
    assert result == {"retcode": 0, "task_type": ""}

    # Check ConfigManager constructor call
    mock_config_manager.assert_called_once()

    # Check ConfigManager method calls
    mock_config_manager.return_value.configure_template.assert_called_once_with(model="")
    mock_config_manager.return_value.build_workspace.assert_called_once()
    mock_config_manager.return_value.configure_data_config.assert_called_once_with()

    # Check Builder constructor call
    mock_builder.assert_called_once_with()



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os

from otx.algorithms.segmentation.adapters.mmseg.utils.builder import build_segmentor
from otx.algorithms.common.adapters.mmdeploy.apis import NaiveExporter
from otx.algorithms.common.adapters.mmcv.tasks.exporter import Exporter
from otx.algorithms.segmentation.adapters.mmseg.utils.exporter import SegmentationExporter
from otx.algorithms.common.adapters.mmcv.utils.config_utils import MPAConfig
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.segmentation.test_helpers import (
    DEFAULT_SEG_TEMPLATE_DIR,
)


@e2e_pytest_unit
def test_run(mocker):
    exporter = SegmentationExporter()
    model_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "model.py"))
    model_cfg.work_dir = "/tmp/"
    args = {"precision": "FP32", "model_builder": build_segmentor}
    mocker.patch.object(Exporter, "run", return_value=True)
    returned_value = exporter.run(model_cfg)
    assert "model_builder" in args
    assert returned_value is True


@e2e_pytest_unit
def test_naive_export(mocker):
    exporter = SegmentationExporter()
    data_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "data_pipeline.py"))
    mock_export_ov = mocker.patch.object(NaiveExporter, "export2openvino")
    exporter.naive_export("", build_segmentor, "FP32", data_cfg)
    mock_export_ov.assert_called_once()



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-Lggggcense-Identifier: Apache-2.0
#

import math

import pytest
import torch
from torch.nn import functional as F

from otx.core.ov.ops.activations import (
    ClampV0,
    EluV0,
    ExpV0,
    GeluV7,
    HardSigmoidV0,
    HSigmoidV5,
    HSwishV4,
    MishV4,
    PReluV0,
    ReluV0,
    SeluV0,
    SigmoidV0,
    SoftMaxV0,
    SoftMaxV1,
    SwishV4,
    TanhV0,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestSoftMaxV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = SoftMaxV0("dummy", shape=self.shape, axis=-1)
        output = op(self.input)
        assert torch.equal(output, F.softmax(self.input, dim=op.attrs.axis))
        assert torch.equal(output >= 0, output <= 1)
        assert torch.allclose(output.sum(-1), torch.tensor(1.0))


class TestSoftMaxV1:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = SoftMaxV1("dummy", shape=self.shape, axis=-1)
        output = op(self.input)
        assert torch.equal(output, F.softmax(self.input, dim=op.attrs.axis))
        assert torch.equal(output >= 0, output <= 1)
        assert torch.allclose(output.sum(-1), torch.tensor(1.0))


class TestReluV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = ReluV0("dummy", shape=self.shape)
        output = op(self.input)
        assert torch.equal(output, F.relu(self.input))
        assert all((output >= 0).flatten())


class TestSwishV4:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = SwishV4("dummy", shape=self.shape)
        output = op(self.input, 1.0)
        assert torch.equal(output, self.input * torch.sigmoid(self.input * 1.0))


class TestSigmoidV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = SigmoidV0("dummy", shape=self.shape)
        output = op(self.input)
        assert torch.equal(output, torch.sigmoid(self.input))
        assert torch.equal(output >= 0, output <= 1)


class TestClampV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = ClampV0("dummy", shape=self.shape, min=-0.05, max=0.05)
        output = op(self.input)
        assert torch.equal(output, self.input.clamp(min=op.attrs.min, max=op.attrs.max))
        assert torch.equal(output >= op.attrs.min, output <= op.attrs.max)


class TestPReluV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = PReluV0("dummy", shape=self.shape)
        output = op(self.input, torch.tensor(0.1))
        assert torch.equal(output, F.prelu(self.input, torch.tensor(0.1)))


class TestTanhV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = TanhV0("dummy", shape=self.shape)
        output = op(self.input)
        assert torch.equal(output, F.tanh(self.input))


class TestEluV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = EluV0("dummy", shape=self.shape, alpha=0.1)
        output = op(self.input)
        assert torch.equal(output, F.elu(self.input, alpha=op.attrs.alpha))


class TestSeluV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = SeluV0("dummy", shape=self.shape)
        output = op(self.input, 0.1, 0.1)
        assert torch.equal(output, 0.1 * F.elu(self.input, alpha=0.1))


class TestMishV4:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = MishV4("dummy", shape=self.shape)
        output = op(self.input)
        #  assert torch.equal(output, F.mish(self.input))
        assert torch.equal(output, self.input * F.tanh(F.softplus(self.input)))


class TestHSwishV4:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = HSwishV4("dummy", shape=self.shape)
        output = op(self.input)
        assert torch.equal(output, F.hardswish(self.input))


class TestHSigmoidV5:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = HSigmoidV5("dummy", shape=self.shape)
        output = op(self.input)
        assert torch.equal(output, F.hardsigmoid(self.input))


class TestExpV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = ExpV0("dummy", shape=self.shape)
        output = op(self.input)
        assert torch.equal(output, torch.exp(self.input))


class TestHardSigmoidV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = HardSigmoidV0("dummy", shape=self.shape)
        output = op(self.input, 0.1, 0.1)
        assert torch.equal(
            output,
            torch.maximum(
                torch.zeros_like(self.input),
                torch.minimum(torch.ones_like(self.input), self.input * 0.1 + 0.1),
            ),
        )


class TestGeluV7:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        with pytest.raises(ValueError):
            GeluV7("dummy", shape=self.shape, approximation_mode="dummy")

        op = GeluV7("dummy", shape=self.shape)
        output = op(self.input)
        assert torch.equal(output, F.gelu(self.input))

        op = GeluV7("dummy", shape=self.shape, approximation_mode="tanh")
        output = op(self.input)
        assert torch.equal(
            output,
            self.input
            * 0.5
            * (1 + F.tanh(torch.sqrt(2 / torch.tensor(math.pi)) * (self.input + 0.044715 * self.input**3))),
        )



"""Unit test for otx.algorithms.action.adapters.mmaction.models.backbones.movinet"""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import pytest
import torch
from torch import nn

from otx.algorithms.action.adapters.mmaction.models.backbones.movinet import (
    BasicBneck,
    Conv2dBNActivation,
    ConvBlock3D,
    MoViNet,
    OTXMoViNet,
    SqueezeExcitation,
    TFAvgPool3D,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestConv2dBNActivation:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.layer = Conv2dBNActivation(3, 16, kernel_size=3, padding=1)

    @e2e_pytest_unit
    def test_conv2d_bn_activation_output_shape(self):
        x = torch.Tensor(1, 3, 32, 32)
        output = self.layer(x)
        assert output.shape == (1, 16, 32, 32)

    @e2e_pytest_unit
    def test_conv2d_bn_activation_attributes(self):
        assert self.layer.kernel_size == (3, 3)
        assert self.layer.stride == (1, 1)
        assert self.layer.out_channels == 16


class TestConvBlock3D:
    @e2e_pytest_unit
    def test_conv_block_3d_output_shape(self):
        x = torch.Tensor(1, 3, 32, 32, 32)
        layer = ConvBlock3D(3, 16, kernel_size=(3, 3, 3), tf_like=True, conv_type="3d")
        output = layer(x)
        assert output.shape == (1, 16, 32, 32, 32)

    @e2e_pytest_unit
    @pytest.mark.parametrize("conv_type", ["3d", "2plus1d"])
    def test_conv_block_3d_attributes(self, conv_type):
        layer = ConvBlock3D(3, 16, kernel_size=(3, 3, 3), tf_like=True, conv_type=conv_type)
        assert layer.kernel_size == (3, 3, 3)
        assert layer.stride == (1, 1, 1)
        assert layer.dim_pad == 2
        assert layer.conv_type == conv_type
        assert layer.tf_like


class TestSqueezeExcitation:
    @pytest.fixture
    def se_block(self):
        return SqueezeExcitation(16, nn.ReLU, nn.Sigmoid, conv_type="2plus1d", squeeze_factor=4, bias=True)

    @e2e_pytest_unit
    def test_scale_output_shape(self, se_block):
        x = torch.Tensor(1, 16, 32, 32, 32)
        scale = se_block._scale(x)
        assert scale.shape == (1, 16, 1, 1, 1)

    @e2e_pytest_unit
    def test_forward_output_shape(self, se_block):
        x = torch.Tensor(1, 16, 32, 32, 32)
        output = se_block(x)
        assert output.shape == (1, 16, 32, 32, 32)

    @e2e_pytest_unit
    def test_se_block_attributes(self, se_block):
        assert se_block.fc1.kernel_size == (1, 1, 1)
        assert se_block.fc2.kernel_size == (1, 1, 1)
        assert se_block.fc1.conv_type == "2plus1d"
        assert se_block.fc2.conv_type == "2plus1d"


class TestTFAvgPool3D:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.pool = TFAvgPool3D()

    @e2e_pytest_unit
    def test_tf_avg_pool_output_shape(self):
        x = torch.Tensor(1, 3, 32, 32, 32)
        output = self.pool(x)
        assert output.shape == (1, 3, 32, 16, 16)

    @e2e_pytest_unit
    def test_tf_avg_pool_output_shape_odd(self):
        x = torch.Tensor(1, 3, 31, 31, 31)
        output = self.pool(x)
        assert output.shape == (1, 3, 31, 16, 16)

    @e2e_pytest_unit
    def test_tf_avg_pool_output_shape_odd_padding(self):
        x = torch.Tensor(1, 3, 30, 30, 30)
        output = self.pool(x)
        assert output.shape == (1, 3, 30, 15, 15)


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


class TestBasicBneck:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.config = AttrDict(
            input_channels=64,
            expanded_channels=64,
            out_channels=64,
            kernel_size=(3, 3, 3),
            padding=(1, 1, 1),
            stride=(1, 1, 1),
            padding_avg=(1, 1, 1),
        )

    @e2e_pytest_unit
    def test_basic_bneck_output_shape(self):
        module = BasicBneck(self.config, tf_like=False, conv_type="3d", activation_layer=nn.ReLU)
        x = torch.randn(1, self.config.input_channels, 32, 32, 32)
        output = module(x)
        assert output.shape == (1, self.config.out_channels, 32, 32, 32)


class TestMoViNet:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.cfg = AttrDict()
        self.cfg.conv1 = AttrDict(
            {
                "input_channels": 3,
                "out_channels": 16,
                "kernel_size": (3, 5, 5),
                "stride": (1, 1, 1),
                "padding": (1, 2, 2),
            }
        )
        self.cfg.blocks = [
            [
                AttrDict(
                    {
                        "input_channels": 16,
                        "expanded_channels": 24,
                        "out_channels": 24,
                        "kernel_size": (3, 3, 3),
                        "stride": (1, 1, 1),
                        "padding": (1, 1, 1),
                    }
                ),
            ]
        ]
        self.cfg.conv7 = AttrDict(
            {
                "input_channels": 40,
                "out_channels": 256,
                "kernel_size": (1, 1, 1),
                "stride": (1, 1, 1),
                "padding": (0, 0, 0),
            }
        )

    @e2e_pytest_unit
    def test_movinet_output_shape(self):
        module = MoViNet(self.cfg)
        x = torch.randn(1, 3, 32, 32, 32)
        module.conv1 = nn.Identity()
        module.blocks = nn.Identity()
        module.conv7 = nn.Identity()
        output = module(x)
        assert output.shape == (1, 3, 1, 1, 1)

    @e2e_pytest_unit
    def test_init_weights(self):
        module = MoViNet(self.cfg)
        module.apply(module._init_weights)
        for m in module.modules():
            if isinstance(m, nn.Conv3d):
                if m.bias is not None:
                    assert m.bias.mean().item() == pytest.approx(0, abs=1e-2)
                    assert m.bias.std().item() == pytest.approx(0, abs=1e-2)
            elif isinstance(m, (nn.BatchNorm3d, nn.BatchNorm2d, nn.GroupNorm)):
                assert m.bias.mean().item() == pytest.approx(0, abs=1e-2)
                assert m.bias.std().item() == pytest.approx(0, abs=1e-2)
            elif isinstance(m, nn.Linear):
                assert m.bias.mean().item() == pytest.approx(0, abs=1e-2)
                assert m.bias.std().item() == pytest.approx(0, abs=1e-2)

    @e2e_pytest_unit
    def test_OTXMoViNet(self):
        model = OTXMoViNet()
        input_tensor = torch.randn(1, 3, 32, 224, 224)
        output_tensor = model(input_tensor)
        assert output_tensor.shape == torch.Size([1, 480, 1, 1, 1])



"""Task of OTX Classification using mmclassification training backend."""

# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import glob
import os
import time
from contextlib import nullcontext
from copy import deepcopy
from functools import partial
from typing import Any, Dict, Optional, Union

import torch
from mmcls.apis import train_model
from mmcls.datasets import build_dataloader, build_dataset
from mmcls.utils import collect_env
from mmcv.runner import wrap_fp16_model
from mmcv.utils import Config, ConfigDict

from otx.algorithms.classification.adapters.mmcls.utils.exporter import (
    ClassificationExporter,
)
from otx.algorithms.classification.task import OTXClassificationTask
from otx.algorithms.common.adapters.mmcv.hooks.recording_forward_hook import (
    ActivationMapHook,
    BaseRecordingForwardHook,
    EigenCamHook,
    FeatureVectorHook,
    ReciproCAMHook,
)
from otx.algorithms.common.adapters.mmcv.utils import (
    adapt_batch_size,
    build_data_parallel,
    get_configs_by_pairs,
    patch_data_pipeline,
)
from otx.algorithms.common.adapters.mmcv.utils import (
    build_dataloader as otx_build_dataloader,
)
from otx.algorithms.common.adapters.mmcv.utils import build_dataset as otx_build_dataset
from otx.algorithms.common.adapters.mmcv.utils.config_utils import (
    MPAConfig,
    update_or_add_custom_hook,
)
from otx.algorithms.common.configs.training_base import TrainType
from otx.algorithms.common.tasks.nncf_task import NNCFBaseTask
from otx.algorithms.common.utils import set_random_seed
from otx.algorithms.common.utils.data import get_dataset
from otx.algorithms.common.utils.logger import get_logger
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.explain_parameters import ExplainParameters
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.model import ModelPrecision
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.core.data import caching
from otx.core.data.noisy_label_detection import LossDynamicsTrackingHook

from .configurer import (
    ClassificationConfigurer,
    IncrClassificationConfigurer,
    SemiSLClassificationConfigurer,
)
from .utils import build_classifier

logger = get_logger()

# TODO Remove unnecessary pylint disable
# pylint: disable=too-many-lines


class MMClassificationTask(OTXClassificationTask):
    """Task class for OTX classification using mmclassification training backend."""

    # pylint: disable=too-many-instance-attributes
    def __init__(self, task_environment: TaskEnvironment, output_path: Optional[str] = None):
        super().__init__(task_environment, output_path)
        self._data_cfg: Optional[Config] = None
        self._recipe_cfg: Optional[Config] = None

    # pylint: disable=too-many-locals, too-many-branches, too-many-statements
    def _init_task(self, export: bool = False):  # noqa
        """Initialize task."""

        if self._multilabel:
            cfg_path = os.path.join(self._model_dir, "model_multilabel.py")
        elif self._hierarchical:
            cfg_path = os.path.join(self._model_dir, "model_hierarchical.py")
        else:
            cfg_path = os.path.join(self._model_dir, "model.py")
        self._recipe_cfg = MPAConfig.fromfile(cfg_path)
        self._recipe_cfg.domain = self._task_type.domain
        self._recipe_cfg.model.multilabel = self._multilabel
        self._recipe_cfg.model.hierarchical = self._hierarchical
        if self._hierarchical:
            self._recipe_cfg.model.head.hierarchical_info = self._hierarchical_info
        self._config = self._recipe_cfg

        set_random_seed(self._recipe_cfg.get("seed", 5), logger, self._recipe_cfg.get("deterministic", False))

        # Belows may go to the configure function
        patch_data_pipeline(self._recipe_cfg, self.data_pipeline_path)

        if not export:
            self._recipe_cfg.merge_from_dict(self._init_hparam())

        if "custom_hooks" in self.override_configs:
            override_custom_hooks = self.override_configs.pop("custom_hooks")
            for override_custom_hook in override_custom_hooks:
                update_or_add_custom_hook(self._recipe_cfg, ConfigDict(override_custom_hook))
        if len(self.override_configs) > 0:
            logger.info(f"before override configs merging = {self._recipe_cfg}")
            self._recipe_cfg.merge_from_dict(self.override_configs)
            logger.info(f"after override configs merging = {self._recipe_cfg}")

        # add Cancel training hook
        update_or_add_custom_hook(
            self._recipe_cfg,
            ConfigDict(type="CancelInterfaceHook", init_callback=self.on_hook_initialized),
        )
        if self._time_monitor is not None:
            update_or_add_custom_hook(
                self._recipe_cfg,
                ConfigDict(
                    type="OTXProgressHook",
                    time_monitor=self._time_monitor,
                    verbose=True,
                    priority=71,
                ),
            )
        self._recipe_cfg.log_config.hooks.append({"type": "OTXLoggerHook", "curves": self._learning_curves})

        # Update recipe with caching modules
        self._update_caching_modules(self._recipe_cfg.data)

        # Loss dynamics tracking
        if getattr(self._hyperparams.algo_backend, "enable_noisy_label_detection", False):
            LossDynamicsTrackingHook.configure_recipe(self._recipe_cfg, self._output_path)

    # pylint: disable=too-many-arguments
    def configure(
        self,
        training=True,
        subset="train",
        ir_options=None,
    ):
        """Patch mmcv configs for OTX classification settings."""

        # deepcopy all configs to make sure
        # changes under MPA and below does not take an effect to OTX for clear distinction
        recipe_cfg = deepcopy(self._recipe_cfg)
        data_cfg = deepcopy(self._data_cfg)
        assert recipe_cfg is not None, "'recipe_cfg' is not initialized."

        if self._data_cfg is not None:
            data_classes = [label.name for label in self._labels]
        else:
            data_classes = None
        model_classes = [label.name for label in self._model_label_schema]

        recipe_cfg.work_dir = self._output_path
        recipe_cfg.resume = self._resume

        if self._train_type == TrainType.Incremental:
            configurer = IncrClassificationConfigurer()
        elif self._train_type == TrainType.Semisupervised:
            configurer = SemiSLClassificationConfigurer()
        else:
            configurer = ClassificationConfigurer()

        options_for_patch_datasets = {"type": "OTXClsDataset", "empty_label": self._empty_label}
        options_for_patch_evaluation = {"task": "normal"}
        if self._multilabel:
            options_for_patch_datasets["type"] = "OTXMultilabelClsDataset"
            options_for_patch_evaluation["task"] = "multilabel"
        elif self._hierarchical:
            options_for_patch_datasets["type"] = "OTXHierarchicalClsDataset"
            options_for_patch_datasets["hierarchical_info"] = self._hierarchical_info
            options_for_patch_evaluation["task"] = "hierarchical"
        elif self._selfsl:
            options_for_patch_datasets["type"] = "SelfSLDataset"

        cfg = configurer.configure(
            recipe_cfg,
            self._model_ckpt,
            data_cfg,
            training,
            subset,
            ir_options,
            data_classes,
            model_classes,
            options_for_patch_datasets=options_for_patch_datasets,
            options_for_patch_evaluation=options_for_patch_evaluation,
        )
        self._config = cfg
        return cfg

    def build_model(
        self,
        cfg: Config,
        fp16: bool = False,
        **kwargs,
    ) -> torch.nn.Module:
        """Build model from model_builder."""
        model_builder = getattr(self, "model_builder", build_classifier)
        model = model_builder(cfg, **kwargs)
        if bool(fp16):
            wrap_fp16_model(model)
        return model

    def _infer_model(
        self,
        dataset: DatasetEntity,
        inference_parameters: Optional[InferenceParameters] = None,
    ):
        """Main infer function."""
        self._data_cfg = ConfigDict(
            data=ConfigDict(
                train=ConfigDict(
                    otx_dataset=None,
                    labels=self._labels,
                ),
                test=ConfigDict(
                    otx_dataset=dataset,
                    labels=self._labels,
                ),
            )
        )

        dump_features = True
        dump_saliency_map = not inference_parameters.is_evaluation if inference_parameters else True

        self._init_task()

        cfg = self.configure(False, "test", None)
        logger.info("infer!")

        # Data loader
        mm_dataset = build_dataset(cfg.data.test)
        dataloader = build_dataloader(
            mm_dataset,
            samples_per_gpu=cfg.data.test_dataloader.get("samples_per_gpu", 1),
            workers_per_gpu=cfg.data.test_dataloader.get("workers_per_gpu", 0),
            num_gpus=len(cfg.gpu_ids),
            dist=cfg.distributed,
            seed=cfg.get("seed", None),
            shuffle=False,
        )

        # Model
        model = self.build_model(cfg, fp16=cfg.get("fp16", False))
        model.eval()
        feature_model = model
        model = build_data_parallel(model, cfg, distributed=False)

        # InferenceProgressCallback (Time Monitor enable into Infer task)
        time_monitor = None
        if cfg.get("custom_hooks", None):
            time_monitor = [hook.time_monitor for hook in cfg.custom_hooks if hook.type == "OTXProgressHook"]
            time_monitor = time_monitor[0] if time_monitor else None
        if time_monitor is not None:
            # pylint: disable=unused-argument
            def pre_hook(module, inp):
                time_monitor.on_test_batch_begin(None, None)

            def hook(module, inp, outp):
                time_monitor.on_test_batch_end(None, None)

            model.register_forward_pre_hook(pre_hook)
            model.register_forward_hook(hook)

        if not dump_saliency_map:
            forward_explainer_hook: Union[nullcontext, BaseRecordingForwardHook] = nullcontext()
        else:
            forward_explainer_hook = ReciproCAMHook(feature_model)
        if not dump_features:
            feature_vector_hook: Union[nullcontext, BaseRecordingForwardHook] = nullcontext()
        else:
            feature_vector_hook = FeatureVectorHook(feature_model)

        eval_predictions = []
        feature_vectors = []
        saliency_maps = []

        with feature_vector_hook:
            with forward_explainer_hook:
                for data in dataloader:
                    with torch.no_grad():
                        result = model(return_loss=False, **data)
                    eval_predictions.extend(result)
                if isinstance(feature_vector_hook, nullcontext):
                    feature_vectors = [None] * len(mm_dataset)
                else:
                    feature_vectors = feature_vector_hook.records  # pylint: disable=no-member
                if isinstance(forward_explainer_hook, nullcontext):
                    saliency_maps = [None] * len(mm_dataset)
                else:
                    saliency_maps = forward_explainer_hook.records  # pylint: disable=no-member
                if len(eval_predictions) == 0:
                    eval_predictions = [None] * len(mm_dataset)

        assert len(eval_predictions) == len(feature_vectors) == len(saliency_maps), (
            "Number of elements should be the same, however, number of outputs are "
            f"{len(eval_predictions)}, {len(feature_vectors)}, and {len(saliency_maps)}"
        )

        outputs = dict(
            eval_predictions=eval_predictions,
            feature_vectors=feature_vectors,
            saliency_maps=saliency_maps,
        )
        return outputs

    # pylint: disable=too-many-branches, too-many-statements
    def _train_model(
        self,
        dataset: DatasetEntity,
    ):
        """Train function in MMClassificationTask."""
        self._data_cfg = ConfigDict(data=ConfigDict())

        for cfg_key, subset in zip(
            ["train", "val", "unlabeled"],
            [Subset.TRAINING, Subset.VALIDATION, Subset.UNLABELED],
        ):
            subset = get_dataset(dataset, subset)
            if subset and self._data_cfg is not None:
                self._data_cfg.data[cfg_key] = ConfigDict(
                    otx_dataset=subset,
                    labels=self._labels,
                )

        self._is_training = True

        self._init_task()

        cfg = self.configure(True, "train", None)
        logger.info("train!")

        timestamp = time.strftime("%Y%m%d_%H%M%S", time.localtime())

        # Environment
        logger.info(f"cfg.gpu_ids = {cfg.gpu_ids}, distributed = {cfg.distributed}")
        env_info_dict = collect_env()
        env_info = "\n".join([(f"{k}: {v}") for k, v in env_info_dict.items()])
        dash_line = "-" * 60 + "\n"
        logger.info(f"Environment info:\n{dash_line}{env_info}\n{dash_line}")

        # Data
        datasets = [build_dataset(cfg.data.train)]

        # Metadata
        meta = dict()
        meta["env_info"] = env_info
        meta["seed"] = cfg.get("seed", 5)
        meta["exp_name"] = cfg.work_dir

        # Model
        model = self.build_model(cfg, fp16=cfg.get("fp16", False))
        model.train()

        if cfg.distributed:
            torch.nn.SyncBatchNorm.convert_sync_batchnorm(model)
            if cfg.dist_params.get("linear_scale_lr", False):
                new_lr = len(cfg.gpu_ids) * cfg.optimizer.lr
                logger.info(
                    f"enabled linear scaling rule to the learning rate. \
                    changed LR from {cfg.optimizer.lr} to {new_lr}"
                )
                cfg.optimizer.lr = new_lr

        validate = bool(cfg.data.get("val", None))
        if validate:
            val_dataset = build_dataset(cfg.data.val, dict(test_mode=True))
            val_loader_cfg = Config(
                cfg_dict={
                    "num_gpus": len(cfg.gpu_ids),
                    "dist": cfg.distributed,
                    "round_up": True,
                    "seed": cfg.seed,
                    "shuffle": False,  # Not shuffle by default
                    "sampler_cfg": None,  # Not use sampler by default
                    **cfg.data.get("val_dataloader", {}),
                }
            )
            val_dataloader = build_dataloader(val_dataset, **val_loader_cfg)
            eval_cfg = cfg.get("evaluation", {})
            eval_cfg["by_epoch"] = cfg.runner["type"] != "IterBasedRunner"
            cfg.custom_hooks.append(
                dict(
                    type="DistCustomEvalHook" if cfg.distributed else "CustomEvalHook",
                    dataloader=val_dataloader,
                    priority="ABOVE_NORMAL",
                    **eval_cfg,
                )
            )

        if self._hyperparams.learning_parameters.auto_decrease_batch_size:
            train_func = partial(train_model, meta=deepcopy(meta), model=deepcopy(model), distributed=False)
            adapt_batch_size(train_func, cfg, datasets, isinstance(self, NNCFBaseTask))  # nncf needs eval hooks

        train_model(
            model,
            datasets,
            cfg,
            distributed=cfg.distributed,
            validate=False,  # For using CustomEvalHook
            timestamp=timestamp,
            meta=meta,
        )

        # Save outputs
        output_ckpt_path = os.path.join(cfg.work_dir, "latest.pth")
        best_ckpt_path = glob.glob(os.path.join(cfg.work_dir, "best_*.pth"))
        if len(best_ckpt_path) > 0:
            output_ckpt_path = best_ckpt_path[0]
        return dict(
            final_ckpt=output_ckpt_path,
        )

    # These need to be moved somewhere
    def _update_caching_modules(self, data_cfg: Config) -> None:
        def _find_max_num_workers(cfg: dict):
            num_workers = [0]
            for key, value in cfg.items():
                if key == "workers_per_gpu" and isinstance(value, int):
                    num_workers += [value]
                elif isinstance(value, dict):
                    num_workers += [_find_max_num_workers(value)]

            return max(num_workers)

        def _get_mem_cache_size():
            if not hasattr(self._hyperparams.algo_backend, "mem_cache_size"):
                return 0

            return self._hyperparams.algo_backend.mem_cache_size

        max_num_workers = _find_max_num_workers(data_cfg)
        mem_cache_size = _get_mem_cache_size()

        mode = "multiprocessing" if max_num_workers > 0 else "singleprocessing"
        caching.MemCacheHandlerSingleton.create(mode, mem_cache_size)

        update_or_add_custom_hook(
            self._recipe_cfg,
            ConfigDict(type="MemCacheHook", priority="VERY_LOW"),
        )

    def _explain_model(self, dataset: DatasetEntity, explain_parameters: Optional[ExplainParameters]):
        """Explain function in MMClassificationTask."""
        explainer_hook_selector = {
            "eigencam": EigenCamHook,
            "activationmap": ActivationMapHook,
            "classwisesaliencymap": ReciproCAMHook,
        }

        self._data_cfg = ConfigDict(
            data=ConfigDict(
                train=ConfigDict(
                    otx_dataset=None,
                    labels=self._labels,
                ),
                test=ConfigDict(
                    otx_dataset=dataset,
                    labels=self._labels,
                ),
            )
        )

        self._init_task()
        cfg = self.configure(False, "test", None)
        # Data loader
        mm_dataset = otx_build_dataset(cfg, "test", build_dataset)
        dataloader = otx_build_dataloader(
            mm_dataset,
            cfg,
            "test",
            build_dataloader,
            distributed=False,
            round_up=False,
        )
        model = self.build_model(cfg, fp16=cfg.get("fp16", False))
        model.eval()
        feature_model = model
        model = build_data_parallel(model, cfg, distributed=False)

        # InferenceProgressCallback (Time Monitor enable into Infer task)
        time_monitor = None
        if cfg.get("custom_hooks", None):
            time_monitor = [hook.time_monitor for hook in cfg.custom_hooks if hook.type == "OTXProgressHook"]
            time_monitor = time_monitor[0] if time_monitor else None
        if time_monitor is not None:

            def pre_hook(module, inp):  # pylint: disable=unused-argument
                time_monitor.on_test_batch_begin(None, None)

            def hook(module, inp, outp):  # pylint: disable=unused-argument
                time_monitor.on_test_batch_end(None, None)

            model.register_forward_pre_hook(pre_hook)
            model.register_forward_hook(hook)

        explainer = explain_parameters.explainer if explain_parameters else None
        if explainer is not None:
            explainer_hook = explainer_hook_selector.get(explainer.lower(), None)
        else:
            explainer_hook = None
        if explainer_hook is None:
            raise NotImplementedError("Explainer algorithm not supported!")

        eval_predictions = []
        with explainer_hook(feature_model) as forward_explainer_hook:
            # do inference and record intermediate fmap
            for data in dataloader:
                with torch.no_grad():
                    result = model(return_loss=False, **data)
                eval_predictions.extend(result)
            saliency_maps = forward_explainer_hook.records

        assert len(eval_predictions) == len(saliency_maps), (
            "Number of elements should be the same, however, number of outputs are "
            f"{len(eval_predictions)}, and {len(saliency_maps)}"
        )

        return eval_predictions, saliency_maps

    def _export_model(self, precision, dump_features):
        self._init_task(export=True)

        cfg = self.configure(False, "test", None)

        self._precision[0] = precision
        export_options: Dict[str, Any] = {}
        export_options["deploy_cfg"] = self._init_deploy_cfg(cfg)
        if export_options.get("precision", None) is None:
            assert len(self._precision) == 1
            export_options["precision"] = str(self._precision[0])

        export_options["deploy_cfg"]["dump_features"] = dump_features
        if dump_features:
            output_names = export_options["deploy_cfg"]["ir_config"]["output_names"]
            if "feature_vector" not in output_names:
                output_names.append("feature_vector")
            if export_options["deploy_cfg"]["codebase_config"]["task"] != "Segmentation":
                if "saliency_map" not in output_names:
                    output_names.append("saliency_map")
        export_options["model_builder"] = getattr(self, "model_builder", build_classifier)

        if self._precision[0] == ModelPrecision.FP16:
            export_options["deploy_cfg"]["backend_config"]["mo_options"]["flags"].append("--compress_to_fp16")

        exporter = ClassificationExporter()
        results = exporter.run(
            cfg,
            **export_options,
        )
        return results

    def _init_deploy_cfg(self, cfg: Config) -> Union[Config, None]:
        base_dir = os.path.abspath(os.path.dirname(self._task_environment.model_template.model_template_path))
        deploy_cfg_path = os.path.join(base_dir, "deployment.py")
        deploy_cfg = None
        if os.path.exists(deploy_cfg_path):
            deploy_cfg = MPAConfig.fromfile(deploy_cfg_path)

            def patch_input_preprocessing(deploy_cfg):
                normalize_cfg = get_configs_by_pairs(
                    cfg.data.test.pipeline,
                    dict(type="Normalize"),
                )
                assert len(normalize_cfg) == 1
                normalize_cfg = normalize_cfg[0]

                options = dict(flags=[], args={})
                # NOTE: OTX loads image in RGB format
                # so that `to_rgb=True` means a format change to BGR instead.
                # Conventionally, OpenVINO IR expects a image in BGR format
                # but OpenVINO IR under OTX assumes a image in RGB format.
                #
                # `to_rgb=True` -> a model was trained with images in BGR format
                #                  and a OpenVINO IR needs to reverse input format from RGB to BGR
                # `to_rgb=False` -> a model was trained with images in RGB format
                #                   and a OpenVINO IR does not need to do a reverse
                if normalize_cfg.get("to_rgb", False):
                    options["flags"] += ["--reverse_input_channels"]
                # value must be a list not a tuple
                if normalize_cfg.get("mean", None) is not None:
                    options["args"]["--mean_values"] = list(normalize_cfg.get("mean"))
                if normalize_cfg.get("std", None) is not None:
                    options["args"]["--scale_values"] = list(normalize_cfg.get("std"))

                # fill default
                backend_config = deploy_cfg.backend_config
                if backend_config.get("mo_options") is None:
                    backend_config.mo_options = ConfigDict()
                mo_options = backend_config.mo_options
                if mo_options.get("args") is None:
                    mo_options.args = ConfigDict()
                if mo_options.get("flags") is None:
                    mo_options.flags = []

                # already defiend options have higher priority
                options["args"].update(mo_options.args)
                mo_options.args = ConfigDict(options["args"])
                # make sure no duplicates
                mo_options.flags.extend(options["flags"])
                mo_options.flags = list(set(mo_options.flags))

            def patch_input_shape(deploy_cfg):
                resize_cfg = get_configs_by_pairs(
                    cfg.data.test.pipeline,
                    dict(type="Resize"),
                )
                assert len(resize_cfg) == 1
                resize_cfg = resize_cfg[0]
                size = resize_cfg.size
                if isinstance(size, int):
                    size = (size, size)
                assert all(isinstance(i, int) and i > 0 for i in size)
                # default is static shape to prevent an unexpected error
                # when converting to OpenVINO IR
                deploy_cfg.backend_config.model_inputs = [ConfigDict(opt_shapes=ConfigDict(input=[1, 3, *size]))]

            patch_input_preprocessing(deploy_cfg)
            if not deploy_cfg.backend_config.get("model_inputs", []):
                patch_input_shape(deploy_cfg)

        return deploy_cfg

    def _init_hparam(self) -> dict:
        params = self._hyperparams.learning_parameters
        warmup_iters = int(params.learning_rate_warmup_iters)
        if self._multilabel:
            # hack to use 1cycle policy
            lr_config = ConfigDict(max_lr=params.learning_rate, warmup=None)
        else:
            lr_config = (
                ConfigDict(warmup_iters=warmup_iters) if warmup_iters > 0 else ConfigDict(warmup_iters=0, warmup=None)
            )

        early_stop = False
        if self._recipe_cfg is not None:
            if params.enable_early_stopping and self._recipe_cfg.get("evaluation", None):
                early_stop = ConfigDict(
                    start=int(params.early_stop_start),
                    patience=int(params.early_stop_patience),
                    iteration_patience=int(params.early_stop_iteration_patience),
                )

        if self._recipe_cfg.runner.get("type").startswith("IterBasedRunner"):  # type: ignore
            runner = ConfigDict(max_iters=int(params.num_iters))
        else:
            runner = ConfigDict(max_epochs=int(params.num_iters))

        config = ConfigDict(
            optimizer=ConfigDict(lr=params.learning_rate),
            lr_config=lr_config,
            early_stop=early_stop,
            data=ConfigDict(
                samples_per_gpu=int(params.batch_size),
                workers_per_gpu=int(params.num_workers),
            ),
            runner=runner,
        )

        if self._train_type.value == "Semisupervised":
            unlabeled_config = ConfigDict(
                data=ConfigDict(
                    unlabeled_dataloader=ConfigDict(
                        samples_per_gpu=int(params.unlabeled_batch_size),
                        workers_per_gpu=int(params.num_workers),
                    )
                )
            )
            config.update(unlabeled_config)
        return config

    # This should be removed
    def update_override_configurations(self, config):
        """Update override_configs."""
        logger.info(f"update override config with: {config}")
        config = ConfigDict(**config)
        self.override_configs.update(config)



"""OTX Algorithms - Detection."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

MMDET_AVAILABLE = True

try:
    import mmdet  # noqa: F401
except ImportError:
    MMDET_AVAILABLE = False



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from copy import deepcopy

import pytest
import torch

from otx.algorithms.segmentation.adapters.mmseg.models import DetConHead
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestDetConHead:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.detcon_head = DetConHead(
            predictor=dict(
                type="SelfSLMLP",
                in_channels=4,
                hid_channels=8,
                out_channels=4,
                norm_cfg=dict(type="BN1d", requires_grad=True),
                with_avg_pool=False,
            ),
            loss_cfg=dict(type="DetConLoss", temperature=0.1),
        )

    @e2e_pytest_unit
    def test_init(self):
        assert self.detcon_head.predictor.__class__.__name__ == "SelfSLMLP"
        assert self.detcon_head.detcon_loss.__class__.__name__ == "DetConLoss"

    @e2e_pytest_unit
    def test_init_weights(self):
        self.detcon_head.predictor.mlp[0].weight = torch.nn.Parameter(
            torch.ones_like(self.detcon_head.predictor.mlp[0].weight)
        )
        old_weights = deepcopy(self.detcon_head.predictor.mlp[0].weight)

        self.detcon_head.init_weights()

        new_weights = self.detcon_head.predictor.mlp[0].weight

        assert torch.all(old_weights != new_weights)

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "projs,projs_tgt,ids,ids_tgt,batch_size,num_samples",
        [(torch.ones((4, 4)), torch.ones((4, 4)), torch.Tensor([0, 0, 0, 0]), torch.Tensor([1, 1, 1, 1]), 2, 1)],
    )
    def test_forward(
        self,
        projs: torch.Tensor,
        projs_tgt: torch.Tensor,
        ids: torch.Tensor,
        ids_tgt: torch.Tensor,
        batch_size: int,
        num_samples: int,
    ):
        loss = self.detcon_head(projs, projs_tgt, ids, ids_tgt, batch_size, num_samples)

        assert isinstance(loss, dict)
        assert "loss" in loss



"""Base configurer for mmseg config."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import importlib
import json
import os
from typing import Any, Dict, List, Optional

import numpy as np
import torch
from mmcv.runner import CheckpointLoader
from mmcv.utils import Config, ConfigDict
from torch import distributed as dist

from otx.algorithms.common.adapters.mmcv.utils import (
    align_data_config_with_recipe,
    build_dataloader,
    build_dataset,
    patch_adaptive_interval_training,
    patch_default_config,
    patch_early_stopping,
    patch_fp16,
    patch_persistent_workers,
    patch_runner,
)
from otx.algorithms.common.adapters.mmcv.utils.config_utils import (
    recursively_update_cfg,
    remove_custom_hook,
    update_or_add_custom_hook,
)
from otx.algorithms.common.utils.logger import get_logger
from otx.algorithms.segmentation.adapters.mmseg.utils import (
    patch_datasets,
    patch_evaluation,
)

logger = get_logger()


# pylint: disable=too-many-public-methods
class SegmentationConfigurer:
    """Patch config to support otx train."""

    def __init__(self):
        self.task_adapt_type: Optional[str] = None
        self.task_adapt_op: str = "REPLACE"
        self.org_model_classes: List[str] = []
        self.model_classes: List[str] = []
        self.data_classes: List[str] = []

    # pylint: disable=too-many-arguments
    def configure(
        self,
        cfg: Config,
        model_ckpt: str,
        data_cfg: Config,
        training: bool = True,
        subset: str = "train",
        ir_options: Optional[Config] = None,
        data_classes: Optional[List[str]] = None,
        model_classes: Optional[List[str]] = None,
    ) -> Config:
        """Create MMCV-consumable config from given inputs."""
        logger.info(f"configure!: training={training}")

        self.configure_base(cfg, data_cfg, data_classes, model_classes)
        self.configure_device(cfg, training)
        self.configure_ckpt(cfg, model_ckpt)
        self.configure_model(cfg, ir_options)
        self.configure_data(cfg, training, data_cfg)
        self.configure_task(cfg, training)
        self.configure_hook(cfg)
        self.configure_samples_per_gpu(cfg, subset)
        self.configure_fp16_optimizer(cfg)
        self.configure_compat_cfg(cfg)
        return cfg

    def configure_base(
        self,
        cfg: Config,
        data_cfg: Optional[Config],
        data_classes: Optional[List[str]],
        model_classes: Optional[List[str]],
    ) -> None:
        """Basic configuration work for recipe.

        Patchings in this function are handled task level previously
        This function might need to be re-orgianized
        """

        options_for_patch_datasets = {"type": "MPASegDataset"}

        patch_default_config(cfg)
        patch_runner(cfg)
        patch_datasets(
            cfg,
            **options_for_patch_datasets,  # type: ignore
        )  # for OTX compatibility
        patch_evaluation(cfg)  # for OTX compatibility
        patch_fp16(cfg)
        patch_adaptive_interval_training(cfg)
        patch_early_stopping(cfg)
        patch_persistent_workers(cfg)

        if data_cfg is not None:
            align_data_config_with_recipe(data_cfg, cfg)

        # update model config -> model label schema
        cfg["model_classes"] = model_classes
        if data_classes is not None:
            train_data_cfg: Config = self.get_data_cfg(data_cfg, "train")
            train_data_cfg["data_classes"] = data_classes
            new_classes: List[str] = np.setdiff1d(data_classes, model_classes).tolist()
            train_data_cfg["new_classes"] = new_classes

    def configure_model(
        self,
        cfg: Config,
        ir_options: Optional[Config],
    ) -> None:
        """Patch config's model.

        Change model type to super type
        Patch for OMZ backbones
        """
        if ir_options is None:
            ir_options = {"ir_model_path": None, "ir_weight_path": None, "ir_weight_init": False}

        cfg.model_task = cfg.model.pop("task", "segmentation")
        if cfg.model_task != "segmentation":
            raise ValueError(f"Given cfg ({cfg.filename}) is not supported by segmentation recipe")

        super_type = cfg.model.pop("super_type", None)
        if super_type:
            cfg.model.arch_type = cfg.model.type
            cfg.model.type = super_type

        # OV-plugin
        ir_model_path = ir_options.get("ir_model_path")
        if ir_model_path:

            def is_mmov_model(key: str, value: Any) -> bool:
                if key == "type" and value.startswith("MMOV"):
                    return True
                return False

            ir_weight_path = ir_options.get("ir_weight_path", None)
            ir_weight_init = ir_options.get("ir_weight_init", False)
            recursively_update_cfg(
                cfg,
                is_mmov_model,
                {"model_path": ir_model_path, "weight_path": ir_weight_path, "init_weight": ir_weight_init},
            )

    def configure_data(
        self,
        cfg: Config,
        training: bool,
        data_cfg: Optional[Config],
    ) -> None:
        """Patch cfg.data.

        Merge cfg and data_cfg
        Wrap original dataset type to MPAsegDataset
        """
        if data_cfg:
            cfg.merge_from_dict(data_cfg)

        train_data_cfg = self.get_data_cfg(cfg, "train")
        for mode in ["train", "val", "test"]:
            if train_data_cfg.type == "MPASegDataset" and cfg.data.get(mode, False):
                if cfg.data[mode]["type"] != "MPASegDataset":
                    # Wrap original dataset config
                    org_type = cfg.data[mode]["type"]
                    cfg.data[mode]["type"] = "MPASegDataset"
                    cfg.data[mode]["org_type"] = org_type

    def configure_task(
        self,
        cfg: Config,
        training: bool,
    ) -> None:
        """Patch config to support training algorithm."""
        if "task_adapt" in cfg:
            logger.info(f"task config!!!!: training={training}")

            # Task classes
            self.configure_classes(cfg)
            # Ignored mode
            self.configure_ignore(cfg)

    def configure_ignore(self, cfg: Config) -> None:
        """Change to incremental loss (ignore mode)."""
        if cfg.get("ignore", False):
            cfg_loss_decode = ConfigDict(
                type="CrossEntropyLossWithIgnore",
                use_sigmoid=False,
                loss_weight=1.0,
            )

            if "decode_head" in cfg.model:
                cfg.model.decode_head.loss_decode = cfg_loss_decode

    # pylint: disable=too-many-branches
    def configure_classes(self, cfg: Config) -> None:
        """Patch classes for model and dataset."""
        org_model_classes = self.get_model_classes(cfg)
        data_classes = self.get_data_classes(cfg)

        if "background" not in org_model_classes:
            org_model_classes = ["background"] + org_model_classes
        if "background" not in data_classes:
            data_classes = ["background"] + data_classes

        # Model classes
        if self.task_adapt_op == "REPLACE":
            if len(data_classes) == 1:  # 'background'
                model_classes = org_model_classes.copy()
            else:
                model_classes = data_classes.copy()
        elif self.task_adapt_op == "MERGE":
            model_classes = org_model_classes + [cls for cls in data_classes if cls not in org_model_classes]
        else:
            raise KeyError(f"{self.task_adapt_op} is not supported for task_adapt options!")

        cfg.task_adapt.final = model_classes
        cfg.model.task_adapt = ConfigDict(
            src_classes=org_model_classes,
            dst_classes=model_classes,
        )

        # Model architecture
        if "decode_head" in cfg.model:
            decode_head = cfg.model.decode_head
            if isinstance(decode_head, list):
                for head in decode_head:
                    head.num_classes = len(model_classes)
            else:
                decode_head.num_classes = len(model_classes)

            # For SupConDetCon
            if "SupConDetCon" in cfg.model.type:
                cfg.model.num_classes = len(model_classes)

        # Task classes
        self.org_model_classes = org_model_classes
        self.model_classes = model_classes

    # Functions below are come from base stage
    def configure_ckpt(self, cfg: Config, model_ckpt: str) -> None:
        """Patch checkpoint path for pretrained weight.

        Replace cfg.load_from to model_ckpt
        Replace cfg.load_from to pretrained
        Replace cfg.resume_from to cfg.load_from
        """
        if model_ckpt:
            cfg.load_from = self.get_model_ckpt(model_ckpt)
        if cfg.get("resume", False):
            cfg.resume_from = cfg.load_from
        if cfg.get("load_from", None) and cfg.model.backbone.get("pretrained", None):
            cfg.model.backbone.pretrained = None

    @staticmethod
    def get_model_ckpt(ckpt_path: str, new_path: Optional[str] = None) -> str:
        """Get pytorch model weights."""
        ckpt = CheckpointLoader.load_checkpoint(ckpt_path, map_location="cpu")
        if "model" in ckpt:
            ckpt = ckpt["model"]
            if not new_path:
                new_path = ckpt_path[:-3] + "converted.pth"
            torch.save(ckpt, new_path)
            return new_path
        return ckpt_path

    @staticmethod
    def get_model_classes(cfg: Config) -> List[str]:
        """Extract trained classes info from checkpoint file.

        MMCV-based models would save class info in ckpt['meta']['CLASSES']
        For other cases, try to get the info from cfg.model.classes (with pop())
        - Which means that model classes should be specified in model-cfg for
          non-MMCV models (e.g. OMZ models)
        """

        def get_model_meta(cfg: Config) -> Config:
            ckpt_path = cfg.get("load_from", None)
            meta = {}
            if ckpt_path:
                ckpt = CheckpointLoader.load_checkpoint(ckpt_path, map_location="cpu")
                meta = ckpt.get("meta", {})
            return meta

        def read_label_schema(ckpt_path, name_only=True, file_name="label_schema.json"):
            serialized_label_schema = []
            if any(ckpt_path.endswith(extension) for extension in (".xml", ".bin", ".pth")):
                label_schema_path = os.path.join(os.path.dirname(ckpt_path), file_name)
                if os.path.exists(label_schema_path):
                    with open(label_schema_path, encoding="UTF-8") as read_file:
                        serialized_label_schema = json.load(read_file)
            if serialized_label_schema:
                if name_only:
                    all_classes = [labels["name"] for labels in serialized_label_schema["all_labels"].values()]
                else:
                    all_classes = serialized_label_schema
            else:
                all_classes = []
            return all_classes

        classes: List[str] = []
        meta = get_model_meta(cfg)
        # for MPA classification legacy compatibility
        classes = meta.get("CLASSES", [])
        classes = meta.get("classes", classes)
        if classes is None:
            classes = []

        if len(classes) == 0:
            ckpt_path = cfg.get("load_from", None)
            if ckpt_path:
                classes = read_label_schema(ckpt_path)
        if len(classes) == 0:
            classes = cfg.model.pop("classes", cfg.pop("model_classes", []))
        return classes

    def get_data_classes(self, cfg: Config) -> List[str]:
        """Get data classes from train cfg."""
        data_classes: List[str] = []
        train_cfg = self.get_data_cfg(cfg, "train")
        if "data_classes" in train_cfg:
            data_classes = list(train_cfg.pop("data_classes", []))
        elif "classes" in train_cfg:
            data_classes = list(train_cfg.classes)
        return data_classes

    @staticmethod
    def get_data_cfg(cfg: Config, subset: str) -> Config:
        """Get subset's data cfg."""
        assert subset in ["train", "val", "test"], f"Unknown subset:{subset}"
        if "dataset" in cfg.data[subset]:  # Concat|RepeatDataset
            dataset = cfg.data[subset].dataset
            while hasattr(dataset, "dataset"):
                dataset = dataset.dataset
            return dataset
        return cfg.data[subset]

    @staticmethod
    def configure_hook(cfg: Config) -> None:
        """Update cfg.custom_hooks based on cfg.custom_hook_options."""

        def update_hook(opt: Config, custom_hooks: Any, idx: int, hook: Config) -> None:
            """Delete of update a custom hook."""
            if isinstance(opt, dict):
                if opt.get("_delete_", False):
                    # if option include _delete_=True, remove this hook from custom_hooks
                    logger.info(f"configure_hook: {hook['type']} is deleted")
                    del custom_hooks[idx]
                else:
                    logger.info(f"configure_hook: {hook['type']} is updated with {opt}")
                    hook.update(**opt)

        hook_cfg = ConfigDict(type="LoggerReplaceHook")
        update_or_add_custom_hook(cfg, hook_cfg)

        custom_hook_options = cfg.pop("custom_hook_options", {})
        # logger.info(f"configure_hook() {cfg.get('custom_hooks', [])} <- {custom_hook_options}")
        custom_hooks = cfg.get("custom_hooks", [])
        for idx, hook in enumerate(custom_hooks):
            for opt_key, opt in custom_hook_options.items():
                if hook["type"] == opt_key:
                    update_hook(opt, custom_hooks, idx, hook)

    def configure_device(self, cfg: Config, training: bool) -> None:
        """Setting device for training and inference."""
        cfg.distributed = False
        if torch.distributed.is_initialized():
            cfg.gpu_ids = [int(os.environ["LOCAL_RANK"])]
            if training:  # TODO multi GPU is available only in training. Evaluation needs to be supported later.
                cfg.distributed = True
                self.configure_distributed(cfg)
        elif "gpu_ids" not in cfg:
            cfg.gpu_ids = range(1)

        # consider "cuda" and "cpu" device only
        if not torch.cuda.is_available():
            cfg.device = "cpu"
            cfg.gpu_ids = range(-1, 0)
        else:
            cfg.device = "cuda"

    def configure_samples_per_gpu(self, cfg: Config, subset: str) -> None:
        """Settings samples_per_gpu for training and inference."""

        dataloader_cfg = cfg.data.get(f"{subset}_dataloader", ConfigDict())
        samples_per_gpu = dataloader_cfg.get("samples_per_gpu", cfg.data.get("samples_per_gpu", 1))

        data_cfg = self.get_data_cfg(cfg, subset)
        if data_cfg.get("otx_dataset") is not None:
            dataset_len = len(data_cfg.otx_dataset)

            if getattr(cfg, "distributed", False):
                dataset_len = dataset_len // dist.get_world_size()

            # set batch size as a total dataset
            # if it is smaller than total dataset
            if dataset_len < samples_per_gpu:
                dataloader_cfg.samples_per_gpu = dataset_len

            # drop the last batch if the last batch size is 1
            # batch size of 1 is a runtime error for training batch normalization layer
            if subset in ("train", "unlabeled") and dataset_len % samples_per_gpu == 1:
                dataloader_cfg["drop_last"] = True
            else:
                dataloader_cfg["drop_last"] = False

            cfg.data[f"{subset}_dataloader"] = dataloader_cfg

    @staticmethod
    def configure_fp16_optimizer(cfg: Config) -> None:
        """Configure Fp16OptimizerHook and Fp16SAMOptimizerHook."""
        fp16_config = cfg.pop("fp16", None)
        if fp16_config is not None:
            optim_type = cfg.optimizer_config.get("type", "OptimizerHook")
            opts: Config = dict(
                distributed=getattr(cfg, "distributed", False),
                **fp16_config,
            )
            if optim_type == "SAMOptimizerHook":
                opts["type"] = "Fp16SAMOptimizerHook"
            elif optim_type == "OptimizerHook":
                opts["type"] = "Fp16OptimizerHook"
            else:
                # does not support optimizerhook type
                # let mm library handle it
                cfg.fp16 = fp16_config
                opts = dict()
            cfg.optimizer_config.update(opts)

    @staticmethod
    def configure_distributed(cfg: Config) -> None:
        """Patching for distributed training."""
        if hasattr(cfg, "dist_params") and cfg.dist_params.get("linear_scale_lr", False):
            new_lr = len(cfg.gpu_ids) * cfg.optimizer.lr
            logger.info(
                f"enabled linear scaling rule to the learning rate. \
                changed LR from {cfg.optimizer.lr} to {new_lr}"
            )
            cfg.optimizer.lr = new_lr

    @staticmethod
    def configure_compat_cfg(
        cfg: Config,
    ):
        """Modify config to keep the compatibility."""

        def _configure_dataloader(cfg: Config) -> None:
            global_dataloader_cfg: Dict[str, str] = {}
            global_dataloader_cfg.update(
                {
                    k: cfg.data.pop(k)
                    for k in list(cfg.data.keys())
                    if k
                    not in [
                        "train",
                        "val",
                        "test",
                        "unlabeled",
                        "train_dataloader",
                        "val_dataloader",
                        "test_dataloader",
                        "unlabeled_dataloader",
                    ]
                }
            )

            for subset in ["train", "val", "test", "unlabeled"]:
                if subset not in cfg.data:
                    continue
                dataloader_cfg = cfg.data.get(f"{subset}_dataloader", None)
                if dataloader_cfg is None:
                    raise AttributeError(f"{subset}_dataloader is not found in config.")
                dataloader_cfg = Config(cfg_dict={**global_dataloader_cfg, **dataloader_cfg})
                cfg.data[f"{subset}_dataloader"] = dataloader_cfg

        _configure_dataloader(cfg)


class IncrSegmentationConfigurer(SegmentationConfigurer):
    """Patch config to support incremental learning for semantic segmentation."""

    def configure_task(self, cfg: ConfigDict, training: bool) -> None:
        """Patch config to support incremental learning."""
        super().configure_task(cfg, training)

        # TODO: Revisit this part when removing bg label -> it should be 1 because of 'background' label
        if len(set(self.org_model_classes) & set(self.model_classes)) == 1 or set(self.org_model_classes) == set(
            self.model_classes
        ):
            is_cls_incr = False
        else:
            is_cls_incr = True

        # Update TaskAdaptHook (use incremental sampler)
        task_adapt_hook = ConfigDict(
            type="TaskAdaptHook",
            src_classes=self.org_model_classes,
            dst_classes=self.model_classes,
            model_type=cfg.model.type,
            sampler_flag=is_cls_incr,
            efficient_mode=cfg["task_adapt"].get("efficient_mode", False),
        )
        update_or_add_custom_hook(cfg, task_adapt_hook)


class SemiSLSegmentationConfigurer(SegmentationConfigurer):
    """Patch config to support semi supervised learning for semantic segmentation."""

    def configure_data(self, cfg: ConfigDict, training: bool, data_cfg: ConfigDict) -> None:
        """Patch cfg.data."""
        super().configure_data(cfg, training, data_cfg)
        # Set unlabeled data hook
        if training:
            if cfg.data.get("unlabeled", False) and cfg.data.unlabeled.get("otx_dataset", False):
                self.configure_unlabeled_dataloader(cfg)

    def configure_task(self, cfg: ConfigDict, training: bool, **kwargs: Any) -> None:
        """Adjust settings for task adaptation."""
        super().configure_task(cfg, training, **kwargs)

        # Remove task adapt hook (set default torch random sampler)
        remove_custom_hook(cfg, "TaskAdaptHook")

    @staticmethod
    def configure_unlabeled_dataloader(cfg: ConfigDict) -> None:
        """Patch for unlabled dataloader."""

        model_task: Dict[str, str] = {
            "classification": "mmcls",
            "detection": "mmdet",
            "segmentation": "mmseg",
        }  # noqa
        if "unlabeled" in cfg.data:
            task_lib_module = importlib.import_module(f"{model_task[cfg.model_task]}.datasets")
            dataset_builder = getattr(task_lib_module, "build_dataset")
            dataloader_builder = getattr(task_lib_module, "build_dataloader")

            dataset = build_dataset(cfg, "unlabeled", dataset_builder, consume=True)
            unlabeled_dataloader = build_dataloader(
                dataset,
                cfg,
                "unlabeled",
                dataloader_builder,
                distributed=cfg.distributed,
                consume=True,
            )

            custom_hooks = cfg.get("custom_hooks", [])
            updated = False
            for custom_hook in custom_hooks:
                if custom_hook["type"] == "ComposedDataLoadersHook":
                    custom_hook["data_loaders"] = [*custom_hook["data_loaders"], unlabeled_dataloader]
                    updated = True
            if not updated:
                custom_hooks.append(
                    ConfigDict(
                        type="ComposedDataLoadersHook",
                        data_loaders=unlabeled_dataloader,
                    )
                )
            cfg.custom_hooks = custom_hooks



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Generator

import pytest

from .test_suite.pytest_insertions import (
    get_pytest_plugins_from_otx,
    otx_conftest_insertion,
    otx_pytest_addoption_insertion,
)
from .unit.api.fixtures.general import label_schema_example  # noqa: F401

pytest_plugins = get_pytest_plugins_from_otx()  # noqa: F405

otx_conftest_insertion(default_repository_name="otx/training_extensions/")  # noqa: F405


def pytest_addoption(parser):
    otx_pytest_addoption_insertion(parser)  # noqa: F405


@pytest.fixture(scope="session")
def tmp_dir_path(request) -> Generator[Path, None, None]:
    prefix = request.config.getoption("--test-workspace")
    with TemporaryDirectory(prefix=prefix) as tmp_dir:
        yield Path(tmp_dir)


@pytest.fixture(autouse=True)
def set_default_tmp_path(tmp_dir_path: Path) -> Generator[None, None, None]:
    origin_tmp_dir = os.environ.get("TMPDIR", None)
    os.environ["TMPDIR"] = str(tmp_dir_path)
    yield
    if origin_tmp_dir is None:
        del os.environ["TMPDIR"]
    else:
        os.environ["TMPDIR"] = origin_tmp_dir


@pytest.fixture(autouse=True, scope="session")
def manage_tm_config_for_testing():
    # check file existance both 'isip' and 'openvino_telemetry' if not, create it.
    # and backup contents if exist
    cfg_dir = os.path.join(os.path.expanduser("~"), "intel")
    isip_path = os.path.join(cfg_dir, "isip")
    otm_path = os.path.join(cfg_dir, "openvino_telemetry")
    isip_exist = os.path.exists(isip_path)
    otm_exist = os.path.exists(otm_path)

    created_cfg_dir = False

    if not os.path.exists(cfg_dir):
        created_cfg_dir = True
        os.makedirs(cfg_dir)

    isip_backup = None

    if not isip_exist:
        with open(isip_path, "w") as f:
            f.write("0")
    else:
        with open(isip_path, "r") as f:
            isip_backup = f.read()

    otm_backup = None
    if not otm_exist:
        with open(otm_path, "w") as f:
            f.write("0")
    else:
        with open(otm_path, "r") as f:
            otm_backup = f.read()

    yield

    # restore or remove
    if not isip_exist:
        os.remove(isip_path)
    else:
        if isip_backup is not None:
            with open(isip_path, "w") as f:
                f.write(isip_backup)

    if not otm_exist:
        os.remove(otm_path)
    else:
        if otm_backup is not None:
            with open(otm_path, "w") as f:
                f.write(otm_backup)

    if created_cfg_dir:
        os.rmdir(cfg_dir)



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest
import torch

from otx.algorithms.classification.adapters.mmcls.models.heads.conv_head import (
    ConvClsHead,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestConvClsHead:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.model = ConvClsHead(
            num_classes=10,
            in_channels=3,
        )

    @e2e_pytest_unit
    def test_simple_test(self):
        input = torch.randn(10, 3, 1, 1)
        outputs = self.model.simple_test((input,), softmax=True)
        assert isinstance(outputs, list)
        assert len(outputs) == 10
        for output in outputs:
            assert np.isclose(output.sum(), np.array(1.0))

        outputs = self.model.simple_test((input,), post_process=False)
        assert isinstance(outputs, torch.Tensor)

    @e2e_pytest_unit
    def test_forward_train(self):
        input = torch.randn(10, 3, 1, 1)
        gt_label = torch.ones(10, dtype=torch.int64)
        output = self.model.forward_train((input,), gt_label)
        assert "loss" in output



"""MoViNet Recognizer for OTX compatibility."""
# pylint: disable=unused-argument
# Copyright (c) OpenMMLab. All rights reserved.
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import functools

from mmaction.models.builder import RECOGNIZERS
from mmaction.models.recognizers.recognizer3d import Recognizer3D


@RECOGNIZERS.register_module()
class MoViNetRecognizer(Recognizer3D):
    """MoViNet recognizer model framework for OTX compatibility."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Hooks for redirect state_dict load/save
        self._register_state_dict_hook(self.state_dict_hook)
        self._register_load_state_dict_pre_hook(functools.partial(self.load_state_dict_pre_hook, self))

    @staticmethod
    def state_dict_hook(module, state_dict, *args, **kwargs):
        """Redirect model as output state_dict for OTX MoviNet compatibility."""
        for key in list(state_dict.keys()):
            val = state_dict.pop(key)
            if "cls_head" in key:
                key = key.replace("cls_head.", "")
            else:
                key = key.replace("backbone.", "")
            state_dict[key] = val

    @staticmethod
    def load_state_dict_pre_hook(module, state_dict, prefix, *args, **kwargs):
        """Redirect input state_dict to model for OTX model compatibility."""
        for key in list(state_dict.keys()):
            val = state_dict.pop(key)
            if "classifier" in key:
                key = key.replace("classifier", "cls_head.classifier")
            else:
                key = prefix + "backbone." + key[len(prefix) :]
            state_dict[key] = val
